from mlpug.trainers.callbacks.lr_scheduler_wrapper import LRSchedulerWrapperBase

from statistics import mean


class LRSchedulerWrapper(LRSchedulerWrapperBase):

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        state = {}
        for name, scheduler in self._schedulers.items():
            state[name] = scheduler.state_dict()

        return state, True

    def set_state(self, state):
        """
        :param state:
        :return: success (True or False)
        """

        success = True
        for name, scheduler_state in state.items():
            if name not in self._schedulers:
                self._log.error(f"Scheduler {name} not found, unable to set state, skipping ...")
                success = False
                continue

            self._schedulers[name].load_state_dict(scheduler_state)

        return success

    def _exec_schedulers(self, training_iter, model_quality=None):
        for name, scheduler in self._schedulers.items():
            if self._metric_to_monitor:
                scheduler.step(model_quality, training_iter)
            else:
                scheduler.step(training_iter)

        return True

    def _get_current_lr(self):
        """
        :return: dict with learning rate, per optimizer
        """

        current_lr = {}
        for name, optimizer in self.optimizers.items():
            lr = []
            for group in optimizer.param_groups:
                lr.append(group['lr'])

            current_lr[name] = mean(lr)

        return current_lr



from mlpug.mlpug_exceptions import CallbackInvalidException
from mlpug.trainers.callbacks.callback import Callback


class DistributedSamplerManager(Callback):

    def __init__(self, sampler, name="DistributedSamplerManager"):
        super(DistributedSamplerManager, self).__init__(name=name)

        if sampler is None or not hasattr(sampler, 'set_epoch') or not callable(sampler.set_epoch):
            raise CallbackInvalidException(name, "No valid DistributedSampler provided, missing set_epoch method")

        self.sampler = sampler
        
    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        self._set_epoch(start_epoch)

        return True

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """

        self._set_epoch(logs["epoch"])

        return True

    def _set_epoch(self, epoch):
        self._log.debug(f"Set epoch {epoch} for sampler ...")
        self.sampler.set_epoch(epoch)





class MLPugException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f"{self.message} :\n\n{self.__cause__}"


class CallbackInvalidException(MLPugException):
    def __init__(self, callback_name, message=None):
        err_msg = f"Callback {callback_name} is invalid"
        if message:
            err_msg += f" : {message}"

        super().__init__(err_msg)


class TrainerInvalidException(MLPugException):
    def __init__(self, message=None):
        err_msg = "Trainer is invalid"
        if message:
            err_msg += f" : {message}"

        super().__init__(err_msg)


class BatchNotChunkableException(MLPugException):
    def __init__(self, message=None):
        err_msg = "Given batch is not chunkable, ensure that the batch object implements the " \
                  "`__len__` and `__getitem__` methods, and that the `__getitem__` method can handle slices."
        if message:
            err_msg += f" : {message}"

        super().__init__(err_msg)


class StateInvalidException(MLPugException):
    def __init__(self, message=None):
        err_msg = "State invalid, unable to set state."
        if message:
            err_msg += f" : {message}"

        super().__init__(err_msg)


class InvalidParametersException(MLPugException):
    def __init__(self, message=None):
        err_msg = "Invalid parameter(s)."
        if message:
            err_msg += f" : {message}"

        super().__init__(err_msg)



import sys

import datetime

from mlpug.trainers.callbacks.callback import Callback
from mlpug.utils import get_value_at, is_chunkable

import basics.base_utils as _


class LogProgress(Callback):

    def __init__(self, log_period=200, set_names=None, batch_level=True, logs_base_path="current", name="LogProgress"):
        super(LogProgress, self).__init__(name=name)

        self.log_period = log_period
        self.set_names = set_names or ["training"]
        self.batch_level = batch_level
        self.logs_base_path = logs_base_path

        self.metric_level_names = {
            'batch': 'Batch',
            'window_average': "Moving average",
            'dataset': "Computed over dataset",
            'epoch': "Epoch"
        }

    def on_batch_training_completed(self, training_batch, logs):
        if not self.batch_level:
            return True

        success = True
        current = self._get_logs_base(logs)
        batch_step = current["batch_step"]

        has_dataset_level_metrics = False
        for set_name in self.set_names:
            dataset_metrics = get_value_at(f"{set_name}.dataset", current, warn_on_failure=False)
            has_dataset_level_metrics |= type(dataset_metrics) is dict and len(dataset_metrics) > 0

            if has_dataset_level_metrics:
                break

        if batch_step == 0 or batch_step % self.log_period == 0 or has_dataset_level_metrics:
            eta = self._calc_eta(logs)
            average_duration = self._get_average_batch_duration(logs)

            sys.stdout.write('\nEpoch {:d}/{:d} - ETA: {:s}\tBatch {:d}/{:d} '
                             'Average batch training time {:s}\n'.format(current["epoch"],
                                                                         logs["final_epoch"],
                                                                         eta,
                                                                         current["batch_step"],
                                                                         logs["final_batch_step"],
                                                                         average_duration))

            for metric_level in ['batch', 'window_average', 'dataset', 'epoch']:
                self._write_metric_logs(metric_level, logs)
                sys.stdout.write(f'\n')

            sys.stdout.write(f'\n')

        return success

    def on_epoch_completed(self, logs):
        current = self._get_logs_base(logs)
        duration = self._get_epoch_duration(logs)
        sys.stdout.write('\nEpoch {:d}/{:d}\tREADY - Duration {:s}\n'.format(current["epoch"],
                                                                           logs["final_epoch"],
                                                                           duration))
        success = True
        for metric_level in ['window_average', 'dataset', 'epoch']:
            self._write_metric_logs(metric_level, logs)
            sys.stdout.write(f'\n')

        sys.stdout.write(f'\n')

        return success

    def _calc_eta(self, logs):

        current = self._get_logs_base(logs)

        eta_str = None
        try:
            training_params = current["training_params"]

            average_batch_duration = training_params["window_average"]["duration"]
            if average_batch_duration and average_batch_duration > 0:
                batch_step = current["batch_step"]
                final_batch_step = logs["final_batch_step"]
                num_batches_to_go = final_batch_step - batch_step + 1

                eta_seconds = int(round(average_batch_duration * num_batches_to_go))

                eta_str = str(datetime.timedelta(seconds=eta_seconds))
            else:
                eta_str = "[UNKNOWN]"
        except Exception as e:
            _.log_exception(self._log, "Unable to calculate epoch ETA", e)

        return eta_str

    def _get_average_batch_duration(self, logs):
        current = self._get_logs_base(logs)

        duration_str = "[UNKNOWN]"
        try:
            duration = current["training_params"]["window_average"]["duration"]
            if duration and duration > 0.0:
                duration = int(duration*1000)
                duration_str = f"{duration}ms"
        except Exception as e:
            _.log_exception(self._log, "Unable to get average batch duration", e)

        return duration_str

    def _get_epoch_duration(self, logs):
        current = self._get_logs_base(logs)

        duration_str = None
        try:
            epoch_duration = int(round(current["training_params"]["epoch"]["duration"]))
            duration_str = str(datetime.timedelta(seconds=epoch_duration))
        except Exception as e:
            _.log_exception(self._log, "Unable to get epoch duration", e)

        return duration_str

    def _write_metric_logs(self, metric_level, logs):
        metrics_log = ''
        for set_name in self.set_names:
            set_metrics_log = self._create_set_metrics_log_for(set_name, metric_level, logs)
            if set_metrics_log is None:
                continue

            metrics_log += f'{set_name:<15}: {set_metrics_log}.\n'

        if len(metrics_log) > 0:
            sys.stdout.write(f'{self.metric_level_names[metric_level]}:\n')
            sys.stdout.write(metrics_log)

    def _create_set_metrics_log_for(self, set_name, metric_level, logs):
        current = self._get_logs_base(logs)

        key_path = f"{set_name}.{metric_level}"
        metrics = get_value_at(key_path, current, warn_on_failure=False)
        return self._create_log_for(metrics)

    def _create_log_for(self, metrics, base_metric=None, log_depth=0):
        if not _.is_dict(metrics):
            return None

        metric_names = set(metrics.keys())
        # TODO : Make this a library level constant
        skip_metric_names = {"auxiliary_results", "duration"}

        num_metrics = len(metric_names-skip_metric_names)
        if num_metrics < 1:
            return None

        log = "\n"*int(log_depth > 0) + "\t"*log_depth
        if base_metric is not None:
            log += f"{base_metric:<15}: "

        for c, (metric, value) in enumerate(metrics.items()):
            if metric in skip_metric_names:
                continue

            if type(value) is dict:
                log += "\n" + self._create_log_for(value, metric, log_depth+1)
            elif isinstance(value, (float, int)):
                log_format = self._get_log_format(value)
                log += log_format.format(metric, value)
            else:
                log += "[UNKNOWN]"

            if c < num_metrics - 1:
                log += ', '

        return log

    def _get_log_format(self, value):
        if abs(value) < 0.1:
            log_format = "{:<9s} {:.3e}"
        else:
            log_format = "{:<9s} {:>9.3f}"

        return log_format


class BatchSizeLogger(Callback):

    def __init__(self, batch_dimension=1, name="BatchSizeLogger"):
        super().__init__(name=name)

        self._batch_dimension = batch_dimension

    def on_batch_training_start(self, training_batch, logs):
        """

        :param training_batch:
        :param logs:

        :return: success (True or False)
        """

        current = self._get_logs_base(logs)

        current['training_params']['batch']['batch_size'] = len(training_batch) if is_chunkable(training_batch) else \
            training_batch[0].size(self._batch_dimension)

        return True




import torch


def cross_entropy(inp, target):
    """
    Calculate cross entropy per sample

    :param inp: (batch_size, voc_size)
    :param target: (batch_size,)

    :return: cross entropy loss

    loss : (batch_size,) loss, per sample in the batch
    """
    return -torch.log(torch.gather(inp, 1, target.view(-1, 1)).squeeze())


def masked_loss(per_sample_loss, mask, average_loss=True):
    """

    :param per_sample_loss: (seq_length, batch_size)
    :param mask: (seq_length, batch_size)
    :param average_loss: If True, calculates the average loss over the samples, else sums the sample losses

    :return: loss

    loss: scalar (as tensor): average, or summed, loss over all samples

    """
    loss = per_sample_loss.masked_select(mask)

    return loss.mean() if average_loss else loss.sum()



from basics.base import Base
from mlpug.mlpug_exceptions import StateInvalidException, InvalidParametersException


class SlidingWindow(Base):

    def __init__(self, length=None, init_window_values=None, state=None, warn=False, **kwargs):
        """

        Can raise StateInvalidException or InvalidParametersException

        :param length:
        :type length:
        :param init_window_values:
        :type init_window_values:
        :param state:
        :type state:
        :param kwargs:
        :type kwargs:
        """
        super().__init__(**kwargs)

        if state:
            try:
                length = state["length"]
                init_window_values = state["window"]
            except Exception as e:
                raise StateInvalidException() from e

        if not length:
            raise InvalidParametersException("No length value given, unable to instantiate SlidingWindow object.")

        if length % 2 != 1 and warn:
            self._log.warn("Given window length is even, centre will not be exact")

        self.length = length
        self.center_index = self.length // 2

        if hasattr(init_window_values, "__getitem__"):
            self._log.debug("Using given initial values to fill the window as much as possible.")

            l = len(init_window_values)
            l = l if l < self.length else self.length

            self.window = init_window_values[-l:][:]
        else:
            self.window = list()

    def get_state(self):
        return {
            "window": self.window,
            "length": self.length
        }

    def flush(self):
        self.window.clear()

    def center(self):
        if not self.is_filled():
            self._log.error("Window not completely filled yet, unable to get center")
            return None

        return self.window[self.center_index]

    def last(self):
        return self.window[-1]

    def first(self):
        return self.window[0]

    def is_filled(self):
        return len(self.window) >= self.length

    def is_empty(self):
        return len(self.window) == 0

    def slide(self, value):
        if self.is_filled():
            self.window = self.window[1:]

        self.window.append(value)



import os
import argparse
import pickle
import random

import numpy as np

import torch
import torch.nn as nn
from torch import optim
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as PT_DDP

from mlpug.reference.chatbot_tutorial.model_data_generation import \
    IndexedSentencePairsDataset, \
    create_sentence_pairs_collate_fn

from mlpug.reference.chatbot_tutorial.seq2seq import EncoderRNN, LuongAttnDecoderRNN
from mlpug.reference.chatbot_tutorial.training import Seq2SeqTrainModel
from mlpug.reference.chatbot_tutorial.loss import masked_loss

import mlpug.pytorch as mlp

from mlpug.utils import get_value_at

from basics.logging import get_logger

parser = argparse.ArgumentParser(description='Train reference chatbot on processed sentence pair data, as generated by '
                                             'chatbot_data_reference.py, using own framework. Also see training.py.')

parser.add_argument(
    '--experiment-name',
    type=str, required=True,
    help='Name of experiment, used as file name base for checkpoints')

parser.add_argument(
    '--dataset-path',
    type=str, required=True,
    help='Path to dataset files')

parser.add_argument(
    '--base-dataset-filename',
    type=str, required=True,
    help='Shared filename postfix for training and validation set')

parser.add_argument(
    '--training-checkpoint',
    type=str, required=False, default=None,
    help='Path to training checkpoint file')

parser.add_argument('--float16',
                    required=False,
                    action='store_const', const=True,
                    help='Provide this flag to use Mixed Precision Training')

parser.add_argument(
    '--mixed-precision-opt-level',
    type=str, required=False, default="O2",
    help='Set to True to use Mixed Precision Training')

parser.add_argument(
    '--num-gpus',
    type=int, required=False, default=None,
    help='The number of GPUs to use in one process (DataParallel will be used)')

parser.add_argument(
    '--local_rank',
    type=int, required=False, default=None,
    help='Set automatically when running this script in distributed parallel mode')

parser.add_argument(
    '--attention-model',
    type=str, required=False, default='dot',
    help='Attention model type: "dot", "general" or "concat"')

parser.add_argument(
    '--embedding-size',
    type=int, required=False, default=256,
    help='Input embedding size')

parser.add_argument(
    '--encoder-state-size',
    type=int, required=False, default=1600,
    help='Encoder state size')

parser.add_argument(
    '--num-encoder-layers',
    type=int, required=False, default=2,
    help='Encoder state size')

parser.add_argument(
    '--num-decoder-layers',
    type=int, required=False, default=2,
    help='Decoder state size')

parser.add_argument(
    '--dropout',
    type=float, required=False, default=0.35,
    help='Drop out rate')

parser.add_argument(
    '--batch-size',
    type=int, required=False, default=256,
    help='Batch size')

parser.add_argument(
    '--gradient-clipping',
    type=float, required=False, default=50.0,
    help='Gradient clipping threshold')

parser.add_argument(
    '--teacher-forcing-ratio',
    type=float, required=False, default=1.0,
    help='How often teacher forcing should be applied, compared to not applying it')

parser.add_argument(
    '--learning-rate',
    type=float, required=False, default=0.0001,
    help='Base learning rate')

parser.add_argument(
    '--decoder-learning-rate-ratio',
    type=float, required=False, default=5.0,
    help='Decoder LR ratio, such that the decoder LR is ratio * base LR')

parser.add_argument(
    '--num-epochs',
    type=int, required=False, default=50,
    help='Number of epochs to train')

parser.add_argument(
    '--progress-logging-period',
    type=int, required=False, default=200,
    help='Period in global (batch) iterations before we log the training progress again.')

parser.add_argument(
    '--seed',
    type=int, required=False, default=0,
    help='Random seed to use to ensure same random split at each restart')

args = parser.parse_args()

num_gpus = args.num_gpus
use_mixed_precision = args.float16 == True
distributed = args.local_rank is not None
single_process_data_parallel = (num_gpus is not None) and (num_gpus > 1)
use_cuda = torch.cuda.is_available()

is_first_worker = (distributed and args.local_rank == 0) or (not distributed)

logger_name = os.path.basename(__file__)
if distributed:
    logger_name += f" (WORKER {args.local_rank})"

logger = get_logger(logger_name)

if distributed and single_process_data_parallel:
    logger.warn("The num_gpus argument will be ignored in Distributed Data Parallel mode. "
                "Number of GPUs will be defined through --nproc_per_node argument of the "
                "torch.distributed.launch script")
    num_gpus = None
    single_process_data_parallel = False

if distributed and not use_cuda:
    logger.error("Distributed Data Parallel mode can only be used when NVidia GPUs are available")
    exit(-1)

if use_mixed_precision and single_process_data_parallel:
    logger.error("Mixed precision training can't be combined with single process data parallel, "
                 "use distributed data parallel mode instead")
    exit(-1)

if use_mixed_precision:
    from apex import amp

# if is_first_worker:
#     import pydevd
#     pydevd.settrace('192.168.178.8', port=57491, stdoutToServer=True, stderrToServer=True)

seed = args.seed
logger.info(f"Seed : {seed}")
np.random.seed(args.seed)
torch.manual_seed(args.seed)

logger.info(f"Using CUDA? {use_cuda}")
logger.info(f"In Distributed Data Parallel mode ? {distributed}")
logger.info(f"Using single process Data Parallel mode "
            f"(multiple GPUs in one process)? {single_process_data_parallel}\n\n")
logger.info(f"Use mixed precision training? {use_mixed_precision}")

if use_mixed_precision:
    logger.info(f"Mixed precision optimization level : {args.mixed_precision_opt_level}")

if distributed:
    logger.info(f"Distributed Data Parallel mode : Using GPU {args.local_rank} ")
    torch.cuda.set_device(args.local_rank)

    dist.init_process_group(backend='nccl', init_method='env://')

device = torch.device("cuda" if use_cuda else "cpu")

##################################################
#
# [START] Setup
#
##################################################

# ############ Conversations dataset #############
# Default word tokens

PAD_token = 0  # Used for padding short sentences
SOS_token = 1  # Start-of-sentence token
EOS_token = 2  # End-of-sentence token

dataset_path = args.dataset_path
base_dataset_filename = args.base_dataset_filename
logger.info(f"dataset_path : {dataset_path}")
logger.info(f"base_dataset_filename : {base_dataset_filename}")

##################################################

# ############ Model configuration ###############
attn_model = args.attention_model
logger.info(f"Attention model : {attn_model}")

embedding_size = args.embedding_size
logger.info(f"embedding_size : {embedding_size}")

encoder_state_size = args.encoder_state_size
logger.info(f"encoder_state_size : {encoder_state_size}")

num_encoder_layers = args.num_encoder_layers
num_decoder_layers = args.num_decoder_layers
logger.info(f"num_encoder_layers : {num_encoder_layers}")
logger.info(f"num_decoder_layers : {num_decoder_layers}")

dropout = args.dropout
logger.info(f"dropout rate : {dropout}")
##################################################

# ########### Training/optimization ##############
experiment_name = args.experiment_name
logger.info(f"experiment_name: {experiment_name}")

batch_size = args.batch_size
if single_process_data_parallel:
    batch_size = num_gpus * batch_size
    logger.info(f"batch_size (all GPUs combined): {batch_size}")
else:
    logger.info(f"batch_size (per process): {batch_size}")

clip = args.gradient_clipping
willClip = clip > 0.0
logger.info(f"Gradient clipping : {clip} (Will clip? {willClip})")

teacher_forcing_ratio = args.teacher_forcing_ratio
logger.info(f"Teacher forcing ratio : {teacher_forcing_ratio}")

learning_rate = args.learning_rate
logger.info(f"Base learning rate : {learning_rate}")

decoder_learning_ratio = args.decoder_learning_rate_ratio
logger.info(f"Decoder LR ratio : {learning_rate}")

num_epochs = args.num_epochs
logger.info(f"Number of training epochs : {num_epochs}")

progress_logging_period = args.progress_logging_period
logger.info(f"Progress logging period : {progress_logging_period}")

# For Tensorboard
metric_names = {
    'batch.loss': 'cross_entropy',
    'duration.batch': 'training_time',
    'duration.window_average': 'training_time',
    'duration.dataset': 'training_time',
    'batch_size': 'size'
}


##################################################

##################################################
#
# [END] Setup
#
##################################################


def load_sentence_pair_data(sentence_pair_file):
    logger.info(f'\nLoading sentence pair file : {sentence_pair_file}\n')

    with open(sentence_pair_file, 'rb') as f:
        data = pickle.load(f)
        return data['pairs'], data['voc']


def transfer_to_gpu(opt):
    # If you have cuda, configure cuda to call
    for state in opt.state.values():
        for k, v in state.items():
            if isinstance(v, torch.Tensor):
                state[k] = v.cuda()


def dataset_path_for(subset):
    return os.path.join(dataset_path, f"{subset}-{base_dataset_filename}")


# ############## SETUP DATASETS ###################
logger.info('Setup data sets ...')

logger.info('Loading training set ...')
training_dataset, voc = load_sentence_pair_data(dataset_path_for('training'))
logger.info('Loading validation set ...')
validation_dataset, _unused_ = load_sentence_pair_data(dataset_path_for('validation'))

logger.debug(f"Number of sentence pairs in training set: {len(training_dataset)}")
logger.debug(f"Number of sentence pairs in validation set: {len(validation_dataset)}")

training_dataset = IndexedSentencePairsDataset(training_dataset, voc, EOS_token)
validation_dataset = IndexedSentencePairsDataset(validation_dataset, voc, EOS_token)

collate_sentence_pairs = create_sentence_pairs_collate_fn(PAD_token, fixed_sequence_length=40)
# collate_sentence_pairs = create_sentence_pairs_collate_fn(PAD_token)

training_sampler = None
validation_sampler = None
if distributed:
    # TODO : do sampler.set_epoch(), at the beginning of each epoch?
    training_sampler = torch.utils.data.distributed.DistributedSampler(training_dataset)
    validation_sampler = torch.utils.data.distributed.DistributedSampler(validation_dataset)

train_dataset_loader = torch.utils.data.DataLoader(training_dataset,
                                                   batch_size=batch_size,
                                                   shuffle=(training_sampler is None),
                                                   num_workers=3,
                                                   collate_fn=collate_sentence_pairs,
                                                   pin_memory=True,
                                                   sampler=training_sampler)

validation_dataset_loader = torch.utils.data.DataLoader(validation_dataset,
                                                        batch_size=batch_size,
                                                        shuffle=(validation_sampler is None),
                                                        num_workers=3,
                                                        collate_fn=collate_sentence_pairs,
                                                        pin_memory=True,
                                                        sampler=validation_sampler)

##################################################


# ############### BUILD MODEL ####################
logger.info('Building model ...')

# Initialize word embeddings
embedding = nn.Embedding(voc.num_words, embedding_size)

# Initialize encoder & decoder models
encoder = EncoderRNN(encoder_state_size, embedding, num_encoder_layers, dropout)
decoder = LuongAttnDecoderRNN(attn_model, embedding, encoder_state_size, voc.num_words, num_decoder_layers, dropout)

training_model = Seq2SeqTrainModel(encoder, decoder)

logger.info(f"Transfer training model to device {device} ...")
# Use appropriate device
training_model = training_model.to(device)

##################################################

# ############## SETUP TRAINING ##################

# Initialize optimizers
logger.info('Building optimizers ...')
encoder_optimizer = optim.Adam(encoder.parameters(), lr=learning_rate)
decoder_optimizer = optim.Adam(decoder.parameters(), lr=learning_rate * decoder_learning_ratio)

# TODO : Is this required for DDP and mixed procession training?
if use_cuda:
    logger.info(f"Transfer optimizers model to {device} ...")
    transfer_to_gpu(encoder_optimizer)
    transfer_to_gpu(decoder_optimizer)

if use_mixed_precision:
    logger.info(f"Initialize model components and optimizers for mixed precision training ...")
    opt_level = args.mixed_precision_opt_level
    training_model, [encoder_optimizer, decoder_optimizer] = \
        amp.initialize(training_model,
                       [encoder_optimizer, decoder_optimizer],
                       opt_level=opt_level,
                       loss_scale="dynamic",
                       min_loss_scale=1.0)

if single_process_data_parallel:
    logger.info(f"We are using {num_gpus} GPUs in a single process, making training model data parallel ...")
    # TODO : actually use the right number of GPUs
    training_model = nn.DataParallel(training_model, dim=1)
elif distributed:
    training_model = PT_DDP(training_model,
                            device_ids=[args.local_rank],
                            output_device=args.local_rank,
                            dim=1)

logger.info('Prepare training ...')


class TeacherForcingController(mlp.callbacks.Callback):

    def on_epoch_start(self, logs):
        if 'training_settings' not in logs:
            logs['training_settings'] = {}
            logs['evaluate_settings'] = {}

        return True

    def on_batch_training_start(self, training_batch, logs):
        # Determine if we are using teacher forcing this iteration
        use_teacher_forcing = True if random.random() < teacher_forcing_ratio else False

        logs['training_settings']['use_teacher_forcing'] = use_teacher_forcing
        logs['evaluate_settings']['use_teacher_forcing'] = use_teacher_forcing

        return True


class ChatbotTrainer(mlp.trainers.DefaultTrainer):

    def _evaluate_loss(self, batch_data, evaluate_settings=None):

        use_teacher_forcing = get_value_at('use_teacher_forcing', evaluate_settings)
        if use_teacher_forcing is None:
            use_teacher_forcing = True

        padded_input_batch, input_lengths, output_batch, output_mask, max_output_len = batch_data

        padded_input_batch = padded_input_batch.to(device)
        input_lengths = input_lengths.to(device)
        output_batch = output_batch.to(device)
        output_mask = output_mask.to(device)

        batch_size = padded_input_batch.size(1)

        init_decoder_input = torch.tensor([[SOS_token for _ in range(batch_size)]], dtype=torch.long, device=device)

        per_sample_loss = self.training_model(padded_input_batch, input_lengths, init_decoder_input, max_output_len,
                                              output_batch, use_teacher_forcing)

        loss = masked_loss(per_sample_loss, output_mask)

        return loss, None

    def _back_propagate_from(self, loss):
        if not use_mixed_precision:
            super()._back_propagate_from(loss)
        else:
            with amp.scale_loss(loss, self.optimizers.values()) as scaled_loss:
                scaled_loss.backward()

    def _prepare_update_model_parameters(self):
        if not willClip:
            return

        # Clip gradients: gradients are modified in place
        if not use_mixed_precision:
            _ = nn.utils.clip_grad_norm_(self.training_model.parameters(), clip)
        else:
            for optimizer in self.optimizers.values():
                _ = nn.utils.clip_grad_norm_(amp.master_params(optimizer), clip)


def calc_loss(loss, **kwargs):
    if distributed:
        dist.reduce(loss, 0)
        if args.local_rank == 0:
            loss /= float(dist.get_world_size())

    return loss.item()


trainer = ChatbotTrainer([encoder_optimizer, decoder_optimizer],
                         {
                             'embedding': embedding,
                             'encoder': encoder,
                             'decoder': decoder
                         })

callbacks = [TeacherForcingController(name="TeacherForcingController"),
             mlp.callbacks.TestMetricsLogger(validation_dataset_loader,
                                             'validation',
                                             {
                                                 'loss': calc_loss
                                             }),
             ]

if is_first_worker:
    callbacks += [
        mlp.callbacks.BatchSizeLogger(),
        mlp.callbacks.CheckpointManager(metric_to_monitor='validation.window_average.loss',
                                        base_checkpoint_filename=args.experiment_name,
                                        archive_last_model_checkpoint_every=20000),
        mlp.callbacks.LogProgress(log_period=progress_logging_period, set_names=["training", "validation"]),
        mlp.callbacks.AutoTensorboard(experiment_name=experiment_name, dataset_name='training',
                                      metric_names=metric_names),
        mlp.callbacks.AutoTensorboard(experiment_name=experiment_name, dataset_name='validation',
                                      metric_names=metric_names),
        # Batch-level batch duration and batch size
        mlp.callbacks.Tensorboard(['duration.batch', 'batch_size'],
                                  experiment_name=experiment_name,
                                  dataset_name='training_params',
                                  metric_names=metric_names,
                                  ignore_missing_metrics=True),
        # Batch-level average batch duration
        mlp.callbacks.Tensorboard(['duration.window_average'],
                                  experiment_name=experiment_name,
                                  dataset_name='training_params',
                                  metrics_are_averages=True,
                                  metric_names=metric_names,
                                  ignore_missing_metrics=True),
        # Epoch-level epoch duration
        mlp.callbacks.Tensorboard(['duration.dataset'],
                                  experiment_name=experiment_name,
                                  dataset_name='training_params',
                                  batch_level=False,
                                  metric_names=metric_names,
                                  ignore_missing_metrics=True)]

manager = mlp.trainers.TrainingManager(trainer,
                                       train_dataset_loader,
                                       num_epochs=num_epochs,
                                       callbacks=callbacks,
                                       experiment_data=args)

tc_file = args.training_checkpoint
if tc_file:
    # Can we make this whole procedure a part of MLPug?
    logger.info(f"Loading training checkpoint : {tc_file}")
    map_location = None
    if distributed:
        map_location = f'cuda:{args.local_rank}'

    checkpoint = torch.load(tc_file, map_location=map_location)

    manager.set_state(checkpoint)

    if distributed:
        dist.barrier()
    logger.info(f"Ready loading training checkpoint.")

    # Required, else loading a checkpoint can lead to OOMs
    del checkpoint
    torch.cuda.empty_cache()

trainer.set_training_model(training_model)

##################################################

logger.info('Start training ...')
manager.start_training()



import sys

import abc
from enum import Enum

import math

import numpy as np

from mlpug.trainers.callbacks.callback import Callback
from mlpug.mlpug_exceptions import CallbackInvalidException
from mlpug.utils import has_key, SlidingWindow

import basics.base_utils as _

from mlpug.utils import get_value_at, set_value_at


class MetricsLoggingMode(Enum):
    BATCH_AND_WINDOW_AVERAGE_METRICS = "BATCH_AND_WINDOW_AVERAGE_METRICS"
    WINDOW_AVERAGE_METRICS = "WINDOW_AVERAGE_METRICS"
    WHOLE_DATASET_METRICS = "WHOLE_DATASET_METRICS"

    @staticmethod
    def will_log_window_average_metrics(mode):
        return mode in {MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                        MetricsLoggingMode.WINDOW_AVERAGE_METRICS}


class MetricsLoggerBase(Callback, metaclass=abc.ABCMeta):

    def __init__(self,
                 dataset_name,
                 batch_metric_funcs,
                 batch_level,
                 logging_mode,
                 dataset=None,
                 evaluate_settings=None,
                 batch_averaging_window=None,
                 batch_metric_reducer_funcs=None,
                 log_condition_func=None,
                 show_dataset_evaluation_progress=False,
                 name="MetricsLoggerBase",
                 **kwargs):
        """

        Different modes of operation:

        Batch level: After every batch training iteration:
            Calculate metrics over a sliding window
                log metrics per batch and over window   batch_level = True, mode = BATCH_AND_WINDOW_AVERAGE_METRICS
                log metrics over window only            batch_level = True, mode = WINDOW_AVERAGE_METRICS
            Calculate metrics over whole data set       batch_level = True, mode = WHOLE_DATASET_METRICS

        Epoch level: After every training epoch:
            Calculate metrics over whole data set       batch_level = False, mode = WHOLE_DATASET_METRICS


        Using `log_condition_func` it can be used to control if metrics are calculated


        # TODO : update documentation after refactor dd 18042020

        :param dataset_name:
        :type dataset_name:
        :param batch_metric_funcs: A dict with keys representing the metric names (e.g. "loss", "recall", etc.) and
                             the corresponding values are functions to calculate the metric value, or to gather
                             information to calculated a combined/averaged metric value over a window,
                             also see batch_metric_reducer_funcs and batch_averaging_window

                             The functions will be called as follows:

                             metric_func(**kwargs)

                             Where kwargs will contain the following keys:
                             batch, loss, auxiliary_results, evaluate_settings

                             Example batch_metric_funcs dict:

                                 def get_loss(loss, **kwargs):
                                    return loss

                                 batch_metric_funcs = {
                                    'loss': get_loss
                                 }

                             The function can also return a dict with metrics. For instance:

                                 def calc_metrics(loss, **kwargs):
                                    return {
                                        "loss": loss,
                                        "perplexity": math.exp(loss)
                                    }

                                 batch_metric_funcs = {
                                    'metrics': calc_metrics
                                 }

                             Last, the function can also gather data, to be used by the corresponding
                             reducer_func, e.g.:

                                 def get_target_and_predicted(batch, auxiliary_results, **kwargs):
                                        target = batch[1]
                                        predicted = auxiliary_results[0]

                                        return target, predicted

                                 batch_metric_funcs = {
                                    'recall': get_target_and_predicted
                                 }

                                 The corresponding reducer_func could be:

                                def calc_recall(window):
                                    target = []
                                    predicted = []
                                    # append all batch-level data
                                    for t, p in window:
                                        target.append(t)
                                        predicted.append(p)

                                    return recall_score(t, p)

                                 batch_metric_reducer_funcs = {
                                    'recall': calc_recall
                                 }

        :type batch_metric_funcs:
        :param evaluate_settings:
        :type evaluate_settings:
        :param batch_averaging_window: Window length in batches, over which the average metric must be calculated
        :type batch_averaging_window: Int
        :param batch_metric_reducer_funcs:
                             An optional dict with keys representing the metric names (e.g. "loss", "recall", etc.) and
                             the corresponding values are functions to calculate the average metric value,
                             based on metrics, or other data, gathered per batch, also see batch_metric_funcs and
                             batch_averaging_window

                             The functions will be called as follows:

                             reducer_func(window), where window is a list with metrics or other data
                             gathered per batch.

                             See `batch_metric_funcs` for example.

                             The default functions assumes the window to be a list of floats and will average the
                             values in this list

        :type batch_metric_reducer_funcs:
        :param average_only: If True, only results of the batch_metric_reducer_funcs will be logged, default is false
        :type average_only: Boolean
        :param get_loss_and_aux_from_logs: If true, the model loss and the auxiliary results are retrieved from the logs
        :type get_loss_and_aux_from_logs:
        :param show_dataset_evaluation_progress
        :type show_dataset_evaluation_progress
        :param name:
        :type name:
        """
        super().__init__(name=f"{dataset_name} dataset {name}", **kwargs)

        self._dataset = dataset
        self._dataset_name = dataset_name

        self._batch_metric_funcs = batch_metric_funcs
        self._batch_level = batch_level

        self._logging_mode = logging_mode

        self._evaluate_settings = evaluate_settings

        self._batch_averaging_window = batch_averaging_window

        self._batch_metric_reducer_funcs = batch_metric_reducer_funcs or {}

        self._log_condition_func = log_condition_func or (lambda logs, dataset_batch: True)

        self._show_dataset_evaluation_progress = show_dataset_evaluation_progress

        self._name = name

        # Add default metric averaging funcs for metrics that don't have a metric averaging func provided:
        for metric_name in self._batch_metric_funcs.keys():
            if metric_name not in self._batch_metric_reducer_funcs:
                self._batch_metric_reducer_funcs[metric_name] = lambda window: np.nanmean(np.array(window))

        self._metric_windows = {}

        self._validate()

        if not self.instance_valid():
            raise CallbackInvalidException(self._name)

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        return {
                    "metric_windows": {metric_path: window.get_state()
                                       for metric_path, window in self._metric_windows.items()}
               }, True

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """

        if not self._check_state(state):
            self._log.error('Given state is invalid, unable to set state')
            return False

        self._metric_windows = {metric_path: SlidingWindow(state=window_state)
                                for metric_path, window_state in state["metric_windows"].items()}

        return True

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
            self._init_metric_windows()

        return True

    def on_batch_training_completed(self, dataset_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._batch_level:
            return True

        if not self._log_condition_func(logs=logs, dataset_batch=dataset_batch):
            return True

        self._init_logs(logs)

        if self._logging_mode is MetricsLoggingMode.WHOLE_DATASET_METRICS:
            return self._calc_whole_dataset_metrics(logs, f"{self._dataset_name}.dataset")
        else:
            current = self._get_logs_base(logs)

            batch_metrics = {}
            if not self._calc_batch_metric_data_from(dataset_batch, batch_metrics, logs):
                return False

            dataset_batch_logs = current[self._dataset_name]['batch']
            if self._logging_mode is MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS:
                # Merge in new batch level results
                dataset_batch_logs = {**dataset_batch_logs, **batch_metrics}
                current[self._dataset_name]['batch'] = dataset_batch_logs

            metric_names = list(self._batch_metric_funcs.keys())
            metric_paths = self._get_metric_paths(dataset_batch_logs, metric_names=metric_names)

            self._update_metrics_windows_for(metric_paths, dataset_batch_logs)

            # gather all window data
            batch_metrics_lists = {p: s.window for p, s in self._metric_windows.items()}
            if not self._reduce(batch_metrics_lists, current[self._dataset_name]['window_average']):
                return False

            return True

    def on_epoch_completed(self, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._batch_level:
            return True

        if not self._log_condition_func(logs=logs, dataset_batch=None):
            return True

        self._init_logs(logs)

        return self._calc_whole_dataset_metrics(logs, f"{self._dataset_name}.dataset")

    def _init_logs(self, logs):
        current = self._get_logs_base(logs)
        if self._dataset_name not in current:
            current[self._dataset_name] = {}

        dataset_metrics = current[self._dataset_name]
        for type in ['batch', 'window_average', 'dataset']:
            if type not in dataset_metrics:
                dataset_metrics[type] = {}

    def _calc_whole_dataset_metrics(self, logs, log_path):
        metric_names = list(self._batch_metric_funcs.keys())

        if self._show_dataset_evaluation_progress:
            self._log.debug(f"Calculating metrics ({', '.join(metric_names)}) on whole {self._dataset_name} dataset")

        self._dataset_iterator = iter(self._dataset)

        batch_metric_data_lists = {}

        metric_paths = None
        # Loop over all dataset batches, this will fill the sliding windows, finally calculate the averages
        for dataset_batch in self._dataset_iterator:
            batch_metric_data_map = {}
            if not self._calc_batch_metric_data_from(dataset_batch, batch_metric_data_map, logs):
                return False

            if metric_paths is None:
                metric_paths = self._get_metric_paths(batch_metric_data_map, metric_names=metric_names)

            for metric_path in metric_paths:
                batch_metric_data = get_value_at(metric_path, batch_metric_data_map)

                batch_metric_data_list = batch_metric_data_lists[metric_path] \
                    if metric_path in batch_metric_data_lists else None

                if batch_metric_data_list is None:
                    batch_metric_data_list = []
                    batch_metric_data_lists[metric_path] = batch_metric_data_list

                batch_metric_data_list += [batch_metric_data]

            if self._show_dataset_evaluation_progress:
                sys.stdout.write('#')
                sys.stdout.flush()

        if self._show_dataset_evaluation_progress:
            sys.stdout.write('\n')
            sys.stdout.flush()

        current = self._get_logs_base(logs)
        reduced_metrics_log = get_value_at(log_path, current)

        return self._reduce(batch_metric_data_lists, reduced_metrics_log)

    def _init_metric_windows(self, reset=False):
        if reset:
            self._metric_windows = {}

        # Do nothing if metric windows are already available
        self._metric_windows = self._metric_windows or {}

        return True

    def _get_metric_paths(self, metrics, metric_names=None, base_path=None):

        if metric_names is None:
            metric_names = list(metrics.keys())

        metric_paths = []
        for name in metric_names:
            value = metrics[name]

            if base_path is not None:
                current_path = f"{base_path}.{name}"
            else:
                current_path = name

            if type(value) is dict:
                path_list = self._get_metric_paths(value, base_path=current_path)
            else:
                path_list = [current_path]

            metric_paths += path_list

        return metric_paths

    def _calc_batch_metric_data_from(self, batch, batch_metric_data, logs):
        success = True

        # Use custom settings if available, else use default settings
        evaluate_settings = get_value_at('evaluate_settings', logs, warn_on_failure=False)
        if evaluate_settings is None:
            evaluate_settings = self._evaluate_settings

        if self._dataset is None:
            current = self._get_logs_base(logs)
            loss = get_value_at(f"{self._dataset_name}.batch.loss", current)
            auxiliary_results = get_value_at(f"{self._dataset_name}.batch.auxiliary_results",
                                             current,
                                             warn_on_failure=False)
        else:
            loss, auxiliary_results = self._evaluate_loss(batch, evaluate_settings)

        metric_func_args = {
            'batch': batch,
            'loss': loss,
            'auxiliary_results': auxiliary_results,
            'evaluate_settings': evaluate_settings
        }

        for metric_name, batch_metric_func in self._batch_metric_funcs.items():
            try:
                batch_metric_data[metric_name] = batch_metric_func(**metric_func_args)
            except Exception as e:
                _.log_exception(self._log, f"Exception occurred calculating {metric_name} data for "
                                           f"{self._dataset_name} batch", e)
                success = False

        return success

    @abc.abstractmethod
    def _evaluate_loss(self, batch, evaluate_settings=None):
        """
        Always returns loss, auxiliary_results tuple
        :param batch:
        :type batch:
        :param evaluate_settings:
        :type evaluate_settings:
        :return: loss, auxiliary_results
        :rtype: tuple
        """
        pass

    def _update_metrics_windows_for(self, metric_paths, batch_metrics):
        for metric_path in metric_paths:
            metric_value = get_value_at(metric_path, batch_metrics)

            sliding_window = self._metric_windows[metric_path] if metric_path in self._metric_windows else None
            if sliding_window is None:
                self._log.debug(f"Creating sliding window for {metric_path}")

                sliding_window = SlidingWindow(length=self._batch_averaging_window)
                self._metric_windows[metric_path] = sliding_window

            sliding_window.slide(metric_value)

    def _reduce(self, batch_metric_data_lists, reduced_metrics):
        success = True

        for metric_path, batch_metric_data_list in batch_metric_data_lists.items():
            try:
                reducer_func = get_value_at(metric_path, self._batch_metric_reducer_funcs)
                set_value_at(metric_path, reduced_metrics, reducer_func(batch_metric_data_list))
            except Exception as e:
                _.log_exception(self._log, f"Exception occurred reducing {metric_path} for {self._dataset_name} dataset "
                                           f"batch metric data", e)
                success = False

        return success

    def _check_state(self, state):
        state_attributes = ['metric_windows']

        for attr in state_attributes:
            if not has_key(state, attr):
                self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                return False

        return True

    def _check_validity_funcs(self, func_dict, func_type):
        valid = True

        if func_dict is None:
            return False

        for metric_name, f in func_dict.items():
            if not callable(f):
                self._log.error(f"No valid {func_type} function provided for {metric_name}, "
                                f"the {self} will not function")
                valid = False

        return valid

    def _validate(self):
        self._valid = True

        if not hasattr(self._dataset, '__iter__'):
            if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
                self._log.debug("Training batch evaluation results will be used to calculate and log metrics.")
            else:
                self._log.error(f"No valid dataset provided to calculated metrics on, "
                                f"the {self} will not function")
                self._valid = False

        self._valid &= self._check_validity_funcs(self._batch_metric_funcs, "batch metric")
        self._valid &= self._check_validity_funcs(self._batch_metric_reducer_funcs, "batch metric combining")

        if not type(self._batch_level) is bool:
            self._log.error(f"Batch level must be True or False, given value is: {self._batch_level}\n"
                            f"the {self} will not function")
            self._valid = False

        if self._logging_mode not in MetricsLoggingMode:
            self._log.error(f"Metric logging mode unknown ({self._logging_mode})\n"
                            f"the {self} will not function")
            self._valid = False

        if not self._batch_level and self._logging_mode is not MetricsLoggingMode.WHOLE_DATASET_METRICS:
            self._log.error(f"For epoch level metric logging, the metric logging can only be "
                            f"MetricsLoggingMode.WHOLE_DATASET_METRICS. value is :{self._logging_mode}\n"
                            f"the {self} will not function")
            self._valid = False

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode) and \
                (not isinstance(self._batch_averaging_window, int) or
                 self._batch_averaging_window <= 0):
            self._log.error(f"A valid batch processing window is required for "
                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
            self._valid = False

        if self._log_condition_func is not None and not callable(self._log_condition_func):
            self._log.error(f"The log condition function must be callable, "
                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
            self._valid = False


class TrainingMetricsLogger(MetricsLoggerBase):
    """
    By default, gets already calculated loss and auxiliary results from logs

    TODO : needs testing
    TODO : what to do for batch_level=false? Allow usage of window average of metrics to be used for epoch level?
    """

    def __init__(self,
                 batch_metric_funcs,
                 batch_level=True,
                 logging_mode=MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                 name="TrainingMetricsLogger",
                 **kwargs):

        super().__init__(dataset_name='training',
                         batch_metric_funcs=batch_metric_funcs,
                         batch_level=batch_level,
                         logging_mode=logging_mode,
                         name=name,
                         **kwargs)

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
            if self._batch_averaging_window is None:
                self._log.debug(f"Set batch_averaging_window to number of batches per epoch : {num_batches_per_epoch}")
                self._batch_averaging_window = num_batches_per_epoch

            if self._batch_averaging_window == math.inf:
                self._log.error("The batch average window is infinite, unable to calculate window average. "
                                "Please set a finite batch_averaging_window during construction of this "
                                "TrainingMetricsLogger.")
                self._valid = False
                return False

        return self.on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)

    def _evaluate_loss(self, batch, evaluate_settings=None):
        """
        No implementation required, because loss and auxilairy results are usually available for the training set
        """
        pass


class TestMetricsLoggerBase(MetricsLoggerBase, metaclass=abc.ABCMeta):
    """
    Child class must implement _evaluate_loss method
    """

    def __init__(self,
                 dataset,
                 dataset_name,
                 batch_metric_funcs,
                 batch_level=True,
                 logging_mode=None,
                 name="TestMetricsLoggerBase",
                 **kwargs):

        if logging_mode is None:
            logging_mode = MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS \
                if batch_level else MetricsLoggingMode.WHOLE_DATASET_METRICS

        super().__init__(dataset=dataset,
                         dataset_name=dataset_name,
                         batch_metric_funcs=batch_metric_funcs,
                         batch_level=batch_level,
                         logging_mode=logging_mode,
                         name=name,
                         **kwargs)

        self._dataset_iterator = None

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        success = super().on_epoch_start(logs)

        if not success:
            self._log.error('A problem occurred, will not continue executing this hook')
            return success

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode) and \
                self._dataset_iterator is None:
            self._dataset_iterator = iter(self._dataset)

        return True

    def on_batch_training_completed(self, training_batch, logs):
        current = self._get_logs_base(logs)
        training_iter = current["global_iter"]

        if self._logging_mode is MetricsLoggingMode.WHOLE_DATASET_METRICS:
            return super().on_batch_training_completed(training_batch, logs)
        else:
            try:
                dataset_batch = next(self._dataset_iterator)
            except StopIteration:
                self._log.debug(f"[Iteration {training_iter}] Rewinding the {self._dataset_name} data set")
                self._dataset_iterator = iter(self._dataset)
                dataset_batch = next(self._dataset_iterator)

            return super().on_batch_training_completed(dataset_batch, logs)

    def _validate(self):
        if not hasattr(self._dataset, '__iter__'):
            self._log.error(f"The given dataset {str(self._dataset)} is not iterable, the {self} will not function")
            self._valid = False

        if self._batch_averaging_window is None:
            try:
                self._batch_averaging_window = len(self._dataset)
                if self._batch_averaging_window == math.inf:
                    self._log.error(f"The batch average window of the {self._dataset_name} data set is infinite, "
                                    f"unable to calculate window average. Please set a finite batch_processing_window "
                                    f"during construction of this TestMetricsLogger.")
                    self._valid = False
                else:
                    self._log.debug(f"Set batch averaging window to number of batches in "
                                    f"{self._dataset_name} data set : {self._batch_averaging_window}")
            except Exception as e:
                _.log_exception(self._log, f"Unable to assess data set length to set the batch averaging window, "
                                           f"{self} will not function", e)
                self._valid = False

        super()._validate()



import os
import argparse
import pickle

import torch
import torch.nn as nn

from mlpug.reference.chatbot_tutorial.evaluation import GreedySearchDecoder, evaluateInput

from mlpug.reference.chatbot_tutorial.seq2seq import EncoderRNN, LuongAttnDecoderRNN

from basics.logging import get_logger

# import pydevd
# pydevd.settrace('192.168.178.7', port=57491, stdoutToServer=True, stderrToServer=True)

logger_name = os.path.basename(__file__)
logger = get_logger(logger_name)

parser = argparse.ArgumentParser(description='Evaluate trained reference chat bot')

parser.add_argument(
    '--model-checkpoint',
    type=str, help='Path to model checkpoint file')

parser.add_argument(
    '--vocabulary-file',
    type=str, help='Path to file containing the vocabulary')

parser.add_argument(
    '--token-selection-method',
    type=str, help='Selection method for token from output distribution : \'max\' or \'random-sample\'. '
                   'In case of random-sample, you can also set --sampling-temp')
parser.add_argument(
    '--sampling-temp',
    type=float, required=False, default=0.2,
    help='Random sampling temperature, used when --token-selection-method \'random-sample\' is selected.')


args = parser.parse_args()

use_cuda = torch.cuda.is_available()
device = torch.device("cuda" if use_cuda else "cpu")

logger.info(f"Using CUDA? {use_cuda}")

# We need to store the setup with the model checkpoint
##################################################
#
# [START] Setup
#
##################################################

# ############ vocabulary #############
# Default word tokens


PAD_token = 0  # Used for padding short sentences
SOS_token = 1  # Start-of-sentence token
EOS_token = 2  # End-of-sentence token

##################################################

# ############ Model configuration ###############
# attn_model = 'dot'
attn_model = 'general'
# attn_model = 'concat'
encoder_state_size = 1000  # 1920  # 1600  # 500
embedding_size = 256  # encoder_state_size  # 256  # encoder_state_size
encoder_n_layers = 2
decoder_n_layers = 2
##################################################

# ################# Evaluation ###################
max_seq_length = 40
##################################################

##################################################
#
# [END] Setup
#
##################################################


def load_vocabulary(vocabulary_file):
    logger.info(f'Loading vocabulary file : {vocabulary_file}')

    with open(vocabulary_file, 'rb') as f:
        data = pickle.load(f)
        return data['voc']


def random_sample(p):
    return sample(p, args.sampling_temp)


def sample(p, temperature=1.0):
    """
    TODO : make it work for batch_size > 0

    :param p: shape (1, 1, num_tokens)
    :param temperature:
    :return:
    """
    p = torch.pow(p, 1.0 / temperature)
    p_sum = torch.sum(p, dim=2)
    p = p / p_sum[:, :, None]

    token_idx = torch.multinomial(p.view(-1), 1)
    token_p = torch.gather(p, 2, token_idx.view(1, -1, 1))

    token_idx = token_idx.view(1, -1)
    token_p = token_p.view(1, -1)

    return token_p, token_idx


voc = load_vocabulary(args.vocabulary_file)

logger.info(f'Building model ...')
# Initialize word embeddings
embedding = nn.Embedding(voc.num_words, embedding_size)

# Initialize encoder & decoder models
encoder = EncoderRNN(encoder_state_size, embedding, encoder_n_layers, dropout=0)
decoder = LuongAttnDecoderRNN(attn_model, embedding, encoder_state_size, voc.num_words, decoder_n_layers, dropout=0)

logger.info(f'Loading model checkpoint ...')
checkpoint = torch.load(args.model_checkpoint)

encoder.load_state_dict(checkpoint["encoder"])
decoder.load_state_dict(checkpoint["decoder"])

encoder.to(device)
decoder.to(device)

# Set dropout layers to eval mode
encoder.eval()
decoder.eval()

select_token_func = None
logger.info(f'Token selection method : {args.token_selection_method}')
if args.token_selection_method == 'max':
    pass
elif args.token_selection_method == 'random-sample':
    logger.info(f'Sampling temperature : {args.sampling_temp}')
    select_token_func = random_sample

# Initialize search module
searcher = GreedySearchDecoder(encoder, decoder, SOS_token, device, select_token_func)

# Begin chatting (uncomment and run the following line to begin)
evaluateInput(encoder, decoder, searcher, voc, max_seq_length, EOS_token, device)



import torch

from mlpug.trainers.callbacks.metrics_logger import *

from mlpug.utils import is_chunkable


class TestMetricsLogger(TestMetricsLoggerBase):

    def __init__(self, *args, name="TestMetricsLogger", **kwargs):
        super().__init__(*args, name=name, **kwargs)

    def _evaluate_loss(self, batch, evaluate_settings=None):
        if is_chunkable(batch):
            # Get raw batch
            batch = batch[:]

        with torch.no_grad():
            loss, auxiliary_results = self.trainer.evaluate_loss(
                batch,
                inference_mode=True,
                evaluate_settings=evaluate_settings)

        return loss, auxiliary_results



import torch

from functools import reduce
from mlpug.trainers.training import *

from mlpug.mlpug_exceptions import TrainerInvalidException, BatchNotChunkableException
from mlpug.utils import is_chunkable


class PTTrainerMixin:

    def _activate_inference_mode(self, inference_mode):
        if inference_mode:
            self.training_model.eval()
        else:
            self.training_model.train()

    def _get_model_state(self, model, model_name=None):
        return model.state_dict()

    def _get_optimizer_state(self, optimizer, optimizer_name=None):
        return optimizer.state_dict()

    def _set_model_state(self, model, state, model_name=None):
        model.load_state_dict(state)

    def _set_optimizer_state(self, optimizer, state, optimizer_name):
        optimizer.load_state_dict(state)


class Trainer(PTTrainerMixin, TrainerBase):
    pass


class DefaultTrainer(PTTrainerMixin, DefaultTrainerBase):

    def train_on(self, batch_data, training_settings=None):
        """
        Use batch_data to perform a training iteration.

        Optionally uses `batch_chunk_size` to evaluate the loss in chunks.
        If a `batch_chunk_size` was given during construction of the trainer, the gradients are updated by evaluating
        the batch in chunks.

        *Note*
        When using chunked batch processing, the default implementation assumes that the
        loss, calculated over a chunk, is the average of the sample losses.

        :param batch_data: batch_data object to train on (e.g. dict, list, tuple)
                           When `batch_chunk_size` is given, `batch_data` must be an object that implements the
                           `__len__` and `__getitem__` methods. Here the `__getitem__` method must be able to deal
                           with slices.
        :param training_settings: optional training_settings object (usually dict)

        :return: loss, auxiliary_results

        loss : number (e.g. float)
        auxiliary_results : can be anything, e.g dict or list with values or data items
        """

        if not self.instance_valid():
            raise TrainerInvalidException()

        self._reset_gradients()

        loss, auxiliary_results = self._calc_gradients(batch_data, training_settings=training_settings)

        self._prepare_update_model_parameters()

        self._update_model_parameters()

        return loss, auxiliary_results

    def _reset_gradients(self):
        for optimizer in self.get_optimizers().values():
            optimizer.zero_grad()

    def _calc_gradients(self, batch_data, training_settings=None):

        if not self.batch_chunk_size:
            loss, auxiliary_results = self.evaluate_loss(batch_data,
                                                         inference_mode=False,
                                                         evaluate_settings=training_settings)

            self._back_propagate_from(loss)
        else:
            chunk_losses, chunk_aux_results, chunk_lengths = self._calc_gradients_chunked(batch_data, training_settings)

            loss, auxiliary_results = self._combine_chunk_results(chunk_losses, chunk_aux_results, chunk_lengths)

        return loss.item(), auxiliary_results

    def _calc_gradients_chunked(self, batch_data, training_settings=None):
        """
        See `train_on` method.

        This method slices the `batch_data` in slices of size `self.batch_chunk_size`. For each slice the loss is
        calculated and the gradients are updated through back prop.

        return: chunk_losses, chunk_aux_results, chunk_lengths
                All three outputs are lists
        """

        if not is_chunkable(batch_data):
            raise BatchNotChunkableException()

        chunk_losses = []
        chunk_aux_results = []
        chunk_lengths = []

        batch_size = len(batch_data)
        num_chunks = math.ceil(batch_size / self.batch_chunk_size)
        for chunk_idx in range(num_chunks):
            chunk_start = chunk_idx*self.batch_chunk_size
            chunk_end = min((chunk_idx+1)*self.batch_chunk_size, batch_size)

            chunk_len = chunk_end-chunk_start

            chunk = batch_data[chunk_start:chunk_end]

            loss, aux_results = self.evaluate_loss(chunk, inference_mode=False, evaluate_settings=training_settings)

            # loss is assumed to be the average over the sample loss for the chunk
            # Divide through batch size to factor in that this loss is part of a larger batch.
            last_chunk = chunk_idx == (num_chunks-1)
            self._back_propagate_from(chunk_len*loss/batch_size, last_chunk=last_chunk)

            chunk_losses += [loss]
            chunk_aux_results += [aux_results]
            chunk_lengths += [chunk_len]

        return chunk_losses, chunk_aux_results, chunk_lengths

    def _combine_chunk_results(self, chunk_losses, chunk_aux_results, chunk_lengths):
        """
        This default implementation assumes that the loss for a chunk is the average loss of all samples in the chunk.
        There is no specific combination logic to combine the chunk auxiliary results
        """
        loss = reduce(lambda tot, c: tot+(c[1]*c[0]), zip(chunk_losses, chunk_lengths), 0)
        num_samples = reduce(lambda tot, l: tot+l, chunk_lengths, 0)

        loss /= num_samples

        return loss, chunk_aux_results

    def _back_propagate_from(self, loss, last_chunk=False):
        loss.backward()

    def _prepare_update_model_parameters(self):
        pass

    def _update_model_parameters(self):
        for optimizer in self.get_optimizers().values():
            optimizer.step()



#
# This script contains the code, extracted from the chatbot_reference.py script, to parse and process the
# Cornell Movie Dialogs Corpus (provided through commandline argument `--corpus-path`).
#
# The resulting sentence pairs and vocabulary is stored in the given `--output-file`
#

from __future__ import print_function
from __future__ import unicode_literals

import os
import argparse

import pickle

from mlpug.reference.chatbot_tutorial.conversation_parsing import \
    printLines, \
    process_corpus_and_extract_sentence_pairs, \
    loadPrepareData, \
    trimRareWords

# import pydevd
# pydevd.settrace('192.168.178.8', port=57491, stdoutToServer=True, stderrToServer=True)

parser = argparse.ArgumentParser(description='Reference implementation to convert the Cornell Movie Dialog Corpus data '
                                             'in to a data structure we can use.')

parser.add_argument(
    '--corpus-path',
    type=str, required=True,
    help='path to the corpus data')

parser.add_argument(
    '--output-file',
    type=str, required=True,
    help='path to output file')

parser.add_argument(
    '--max-sequence-length',
    type=int, default=10,
    help='Max sequence length')

parser.add_argument(
    '--min-word-count',
    type=int, default=3,
    help='Minimum vocabulary word count')

args = parser.parse_args()


##################################################
#
# [START] Setup
#
##################################################

# ######## Conversations dataset parsing #########
corpus_name = "cornell-movie-dialogs-corpus"
corpus_path = args.corpus_path

sentence_pair_csv_delimiter = '\t'
raw_sentence_pair_filename = 'formatted_movie_lines.txt'

output_file_path = args.output_file

# Default word tokens
PAD_token = 0  # Used for padding short sentences
SOS_token = 1  # Start-of-sentence token
EOS_token = 2  # End-of-sentence token

MAX_LENGTH = args.max_sequence_length  # 50  # 10  # Maximum sentence length to consider

MIN_COUNT = args.min_word_count  # Minimum word count threshold for trimming
##################################################

##################################################
#
# [END] Setup
#
##################################################

# # Load & Preprocess Data
printLines(os.path.join(corpus_path, "movie_lines.txt"))

# ## Extract sentence pairs and save to disk.

# Define path to new file
sentence_pairs_path = os.path.join(corpus_path, raw_sentence_pair_filename)
process_corpus_and_extract_sentence_pairs(corpus_path, sentence_pairs_path, sentence_pair_csv_delimiter)

# Print a sample of lines
print("\nSample lines from file:")
printLines(sentence_pairs_path)

# ## Filter out long sentences and create vocabulary

# Load/Assemble voc and pairs
voc, pairs = loadPrepareData(corpus_name,
                             sentence_pairs_path,
                             MAX_LENGTH,
                             PAD_token, SOS_token, EOS_token)

# Print some pairs to validate
print("\npairs:")
for pair in pairs[:10]:
    print(pair)

# ## Load sentence pairs and remove pairs with rare words

# Trim voc and pairs
pairs = trimRareWords(voc, pairs, MIN_COUNT)

with open(args.output_file, 'wb') as f:
    pickle.dump({
        'pairs': pairs,
        'voc': voc,
        'args': args
    }, f)



import torch

from mlpug.trainers.callbacks.checkpoint_manager import CheckpointManagerBase

import basics.base_utils as _


class CheckpointManager(CheckpointManagerBase):

    def __init__(self, *args, **kwargs):

        ext_args = {
            'model_checkpoint_filename_ext': 'pt',
            'training_checkpoint_filename_ext': 'tar'
        }

        if kwargs is not None:
            for arg, val in ext_args.items():
                if arg in kwargs:
                    continue

                kwargs[arg] = val
        else:
            kwargs = ext_args

        super().__init__(*args, **kwargs)

    def _save_training_checkpoint(self, filename):
        state, success = self.training_manager.get_state()

        if state:
            if not success:
                self._log.warn("Getting the training state was not completely successful, "
                               "trying to save available state data any way ...")

            try:
                self._log.debug(f"Saving training checkpoint : {filename}")
                torch.save(state, filename)
            except Exception as e:
                _.log_exception(self._log, "Saving the training checkpoint failed", e)
                success = False

        return success

    def _save_model_checkpoint(self, filename):
        state, success = self.trainer.get_model_components_state()

        if state:
            if not success:
                self._log.warn("Getting the model components state was not completely successful, "
                               "trying to save available state data any way ...")

            if self._model_hyper_parameters:
                try:
                    state['hyper_parameters'] = self._model_hyper_parameters
                except Exception as e:
                    _.log_exception(self._log, "Unable to add model hyper parameters to model checkpoint, "
                                               "trying to save the checkpoint anyway ...", e)
                    success = False

            try:
                manager_state, manager_state_success = self.training_manager.get_state_for_model_checkpoint()

                if manager_state_success:
                    state['manager_state'] = manager_state
                else:
                    self._log.warn("Getting the manager state for the model checkpoint was not successful, "
                                   "will continue any way without this data ...")
                    success = False

            except Exception as e:
                _.log_exception(self._log, "Unable to add manager state to model checkpoint, "
                                           "trying to save the checkpoint anyway ...", e)
                success = False

            try:
                self._log.debug(f"Saving model checkpoint : {filename}")
                torch.save(state, filename)
            except Exception as e:
                _.log_exception(self._log, "Saving the model checkpoint failed", e)
                success = False

        return success



# TODO : remove PyTorch dependency
import os
from torch.utils.tensorboard import SummaryWriter

from mlpug.trainers.callbacks.callback import Callback
from mlpug.utils import get_value_at as get_value_at_func

import basics.base_utils as _


# Tensorboard writer types
METRIC_WRITER = 'metrics'
WINDOW_AVERAGED_METRICS_WRITER = 'window-averaged-metrics'


def get_value_at(path, obj):
    return get_value_at_func(path, obj, warn_on_failure=False)


class Tensorboard(Callback):

    def __init__(self,
                 metric_paths,
                 experiment_name=None,
                 dataset_name=None,
                 label_name=None,
                 batch_level=True,
                 metrics_are_averages=False,
                 metric_names=None,
                 batch_log_period=1,
                 flush_period=50,
                 log_dir='../training-logs/',
                 tensorboard_options=None,
                 ignore_missing_metrics=False,
                 name="Tensorboard"):
        """

        Shows batch-level or epoch-level metrics, as defined by the metric key paths in `metric_paths`.

        For instance, assume that the `logs` object, as received by the training manager, is as follows:
        ```
        {
            ...
            'duration': {
                'batch': 0.534
                'epoch': 623.23
            },
            ...
            'batch_size': 200
            ...
        }
        ```

        Then, the following `Tensorboard`
        ```
        Tensorboard(['duration.batch', 'batch_size'], 'experiment1')
        ```
        Will show a figure with the batch duration and the batch size for each batch.




        ## Using `dataset_name`

        If `dataset_name` is given, it looks for the given `metric_paths` in the path with `dataset_name` first.
        Only, if the a given path is not available within the `dataset_name` path, we try to retrieve it as is.

        For example, assume that the `logs` object, as received by the training manager, is as follows:
        ```
        {
            ...
            'training': {
                'window_average': {
                    'loss': 0.201
                }
            },
            'validation': {
                'window_average': {
                    'loss': 0.234
                }
            },
            'duration': {
                'batch': 0.534,
                'window_average' : {
                    'batch' : 0.531
                }
            }
        }
        ```
        The following `Tensorboard`
        ```
        Tensorboard(['window_average.loss', 'duration.window_average.batch'], 'experiment1', 'validation')
        ```
        Will display a figure with `validation.window_average.loss`, because `window_average.loss` is
        available in `validation`.


        ### Showing metrics for different datasets in one graph using `dataset_name`

        By creating multiple `Tensorboard` instances with different data set names, using `dataset_name`, a metric
        path, defined in all instances will be shown in one graph.

        ```
        t_train = Tensorboard(['window_average.loss'], 'experiment1', 'training')
        t_val = Tensorboard(['window_average.loss'], 'experiment1', 'validation')
        ```

        Registering the above tensorboards will plot the window_average Loss for the training set and validation set in
        one graph




        ## Using `label_name` and `metric_names`

        Using `label_name`, in combination with `metric_names`, allows us to combine multiple plots for the
        same metric, but for different categories or labels, in one figure.

        for instance, if we have the following data in a log object:
        {
            ...
           'training': {
                'recall': {
                    'hamburger' : 0.646
                    'hotdog': 0.873
                },
                'precision': {
                    'hamburger' : 0.931
                    'hotdog': 0.829
                }
            },
            ...
        }

        ```
        metric_names = {
            'recall.hamburger': 'Recall',
            'recall.hotdog': 'Recall',
            'precision.hamburger': 'Precision',
            'precision.hotdog': 'Precision'
        }
        ```

        we can create the following tensorboard callbacks
        ```
        t_hamburger = Tensorboard(['recall.hamburger', 'precision.hamburger'],
                                   experiment_name = 'experiment1',
                                   dataset_name = 'training',
                                   label_name = 'hamburger',
                                   metric_names = metric_names)
        t_hotdog = Tensorboard(['recall.hotdog', 'precision.hotdog'],
                                   experiment_name = 'experiment1',
                                   dataset_name = 'training',
                                   label_name = 'hotdog',
                                   metric_names = metric_names)
        ```

        When registering these boards the `hamburger` and `hotdog` recall plots will be combined in one recall plot.
        Further, the `hamburger` and `hotdog` precision plots will be combined in one precision plot.



        ## Using `metrics_are_averages` and `metric_names`

        If `metrics_are_averages=True` the metrics are written to Tensorboard using another writer, such that
        a metric and its average can be shown in one figure.

        For instance,

        ```
        metric_names = {
            'batch.loss': 'cross_entropy',
            'window_average.loss': 'cross_entropy'
        }

        Tensorboard([loss], 'experiment1', 'validation', metric_names = metric_names)
        ```
        Displays a figure with the `validation.batch.loss` at every batch as `cross_entropy`

        When also adding the following `Tensorboard`:
        ```
        Tensorboard([window_average.loss],
                    'experiment1',
                    'validation',
                    metric_names = metric_names,
                    metrics_are_averages=True)
        ```

        Adds the `validation.window_average.loss` averages to the same batch-level `cross_entropy` figure.





        To show the Tensorboard start is as follows:

        `[nohup] tensorboard --logdir <log_dir>/<experiment_name> [--port <my port>] [--bind_all] [&]`

        E.g.

        `tensorboard --logdir ../training-logs/my-first-exp-01012020-0`

        Note : only provide the path up to the `<log_dir>/<experiment_name>`

        :param metric_paths: {list} The key paths in the logs object, specifying the metrics to show.

        :param experiment_name: See log_dir

        :param dataset_name:    See log_dir. Further, when provided, the callback will look for metrics in the
                                <dataset_name>.* path first of the logs object provided by the training manager.

        :param dataset_name:    See log_dir. Further, when provided

        :param batch_level:     {boolean} If True, the given metrics will be logged per batch, else per epoch

        :param metrics_are_averages: {boolean}
                                     If True, the given metrics are written by a separate writer dedicated
                                     to metrics window averages

        :param metric_names:    {dict}
                                Dict mapping metric paths (keys), to metric names for the metrics (values).

                                Example:
                                {
                                    'loss': 'cross_entropy',
                                    'special.my_metric': 'my_awesome_metrics'
                                }

        :param batch_log_period: {int}
                                 period in batch iterations, after which the `batch_metrics` are logged

        :param flush_period : {int}
                                 period in batch iterations, after which the tensorboard writers are flushed

        :param tensorboard_options: {None|dict}
                                    Optional dictionary with options that need to be used when instantiating the
                                    Tensorboard writers.

        :param log_dir:         Logging directory. When `experiment_name`, `dataset_name` and `label_name` are
                                provided, the final logging directory will be:

                                log_dir/<experiment_name>/<label_name>-<dataset-name>-[metric or metric-window-averages]

                                Any parameters not provided will not be used to build up the log_dir

        :param ignore_missing_metrics {Boolean}

        :param name:
        """

        super().__init__(name=name)

        self._metric_paths = metric_paths

        self._experiment_name = experiment_name
        self._dataset_name = dataset_name
        self._label_name = label_name

        self._batch_level = batch_level
        self._metrics_are_averages = metrics_are_averages

        self._metric_names = metric_names

        self._batch_log_period = batch_log_period

        self._flush_period = flush_period

        self._log_dir = log_dir

        self._tensorboard_options = tensorboard_options or {}

        self._ignore_missing_metrics = ignore_missing_metrics

        self._writer_type = WINDOW_AVERAGED_METRICS_WRITER if self._metrics_are_averages else METRIC_WRITER

        self._writer = None

    def _setup(self):
        try:
            success = self._setup_writer_for(self._writer_type)
        except Exception as e:
            _.log_exception(self._log, "An exception occurred setting up the Tensorboard callback", e)
            success = False

        self._valid = success

        return self._valid

    def get_dataset_name(self):
        return self._dataset_name

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        return self._setup()

    def on_batch_training_completed(self, dataset_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._batch_level:
            return True

        current = self._get_logs_base(logs)
        global_iter = current['global_iter']

        if global_iter > 0 and (global_iter % self._batch_log_period != 0):
            return True

        return self._write(current, global_iter)

    def on_epoch_completed(self, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._batch_level:
            return True

        current = self._get_logs_base(logs)
        epoch = current['epoch']

        return self._write(current, epoch)

    def on_training_ended(self, stopped_early, stopped_on_error, callback_calls_success):
        self._writer.close()
        return True

    def _setup_tensorboard_log_dir(self, writer_type):
        """
        TODO : Put in TensorBase class so it can be reused by all Tensorboard types implemented

        :param writer_type:
        :return:
        """
        paths = [self._log_dir]
        if self._experiment_name:
            paths.append(self._experiment_name)

        writer_name = []
        if self._label_name:
            writer_name.append(self._label_name)

        if self._dataset_name:
            writer_name.append(self._dataset_name)

        writer_name.append(writer_type)

        writer_name = '-'.join(writer_name)

        paths.append(writer_name)

        tensorboard_log_dir = os.path.join(*paths)

        self._log.debug(f'Tensorboard directory for {writer_type}: {tensorboard_log_dir}')

        if not os.path.exists(tensorboard_log_dir):
            self._log.debug(f'Directory doesn\'t exist, creating directory')
            os.makedirs(tensorboard_log_dir)

        return tensorboard_log_dir

    def _setup_writer_for(self, writer_type):
        """
        TODO : Put in TensorBase class so it can be reused by all Tensorboard types implemented

        :param writer_type:
        :return:
        """
        self._log.debug(f'Setting up Tensorboard writer for {writer_type} ...')

        tensorboard_log_dir = self._setup_tensorboard_log_dir(writer_type)

        tensorboard_options = {**{
            'log_dir': tensorboard_log_dir
        }, **self._tensorboard_options}

        self._writer = SummaryWriter(**tensorboard_options)

        return True

    def _write(self, current_logs, training_iter):
        metrics, success = self._get_metrics_from(current_logs)

        for tag, metric in metrics.items():
            self._writer.add_scalar(tag, metric, global_step=training_iter)

        if training_iter % self._flush_period == 0:
            self._writer.flush()

        return success

    def _get_metrics_from(self, current_logs):
        if not _.is_sequence(self._metric_paths) or len(self._metric_paths) == 0:
            self._log.error(f'No valid metrics to show, nothing to get')
            return None, False

        return self._get_specific_metrics_from(current_logs, self._metric_paths)

    def _get_specific_metrics_from(self, current_logs, metric_paths):
        success = True
        metrics = {}

        def _try_add_metric(metric_path, metric):
            if metric is None:
                return None

            tag = self._get_tag(metric_path)
            metrics[tag] = metric

            return metric

        for metric_path in metric_paths:
            if self._dataset_name:
                set_metric_path = f'{self._dataset_name}.{metric_path}'
                metric = get_value_at(set_metric_path, current_logs)

                if _try_add_metric(metric_path, metric) is not None:
                    continue

            metric = get_value_at(metric_path, current_logs)

            if not _try_add_metric(metric_path, metric) and not self._ignore_missing_metrics:
                self._log.error(f'Unable to find metric for metric path {metric_path}')
                success = False

        return metrics, success

    def _get_tag(self, metric_name):
        prefix = 'batch' if self._batch_level else 'epoch'

        tag = self._metric_names[metric_name] if (metric_name in self._metric_names) else metric_name
        tag = f"{prefix}_{tag}"

        return tag


class AutoTensorboard(Callback):

    def __init__(self,
                 dataset_name,
                 experiment_name=None,
                 show_batch_level=True,
                 show_batch_window_averages=True,
                 show_epoch_level=True,
                 metric_names=None,
                 batch_log_period=1,
                 flush_period=50,
                 log_dir='../training-logs/',
                 tensorboard_options=None,
                 name="AutoTensorboard"):
        """

        Shows batch-level metrics, and/or averaged batch-level (metrics), and/or epoch level metrics.

        All available metrics for `dataset_name` in the `logs` object received from the
        training manager are logged

        Example, if `dataset_name='validation'` and the received `logs` object contains:

        {
            ...

            'validation': {
                            'batch': {
                                'loss': 0.65,
                            },
                            'window_average': {
                                      'loss': 0.63
                                    }
                           }
            ...
        }

        When `show_batch_level` is True, then `validation.loss` will be shown in
        the batch-level `loss` figure.

        When `show_batch_window_averages` is True, then `validation.window_averages.loss` will be shown in
        the batch-level `loss` figure.

        When `show_epoch_level` is True, then `validation.dataset.loss` will be shown in
        the epoch-level `loss` figure. Further, if `validation.dataset.loss` is not available,
        `validation.window_average.loss` will be tried


        To show the Tensorboard start is as follows:

        `[nohup] tensorboard --logdir <log_dir>/<experiment_name> [--port <my port>] [--bind_all] [&]`

        E.g.

        `tensorboard --logdir ../training-logs/my-first-exp-01012020-0`

        Note : only provide the path up to `<log_dir>/<experiment_name>`

        :param experiment_name: Experiment name, used in the log_dir, see log_dir.

        :param dataset_name:    See log_dir. Further, when provided, the callback will look for metrics in the
                                <dataset_name>.* path first of the logs object provided by the training manager.

        :param show_batch_level: {Boolean} If True, batch-level metrics will be logged to Tensorboard
        :param show_batch_window_averages: {Boolean} If True, (sliding-window) averaged batch-level metrics will be logged
                                           to Tensorboard
        :param show_epoch_level: {Boolean} If True, epoch-level metrics will be logged to Tensorboard

        :param metric_names:    {dict}
                                Dict mapping metric paths (keys), to metric names for the metrics (values).

                                Example:
                                {
                                    'loss': 'cross_entropy',
                                    'special.my_metric': 'my_awesome_metrics'
                                }

                                In this case `validation.loss` and `validation.window_average.loss` are
                                named 'cross_entropy'

        :param batch_log_period: {int}
                                 period in batch iterations, after which the `batch_metrics` are logged

        :param flush_period : {int}
                                 period in batch iterations, after which the tensorboard writers are flushed

        :param log_dir:         Logging directory. When experiment_name and dataset_name are provided, the final
                                logging directory will be:

                                log_dir/experiment_name/dataset_name

                                When only the dataset_name is given:

                                log_dir/dataset_name

        :param tensorboard_options: {None|dict}
                                    Optional dictionary with options that need to be used when instantiating the
                                    Tensorboard writers.

        :param name:
        """
        super().__init__(name=name)

        self._experiment_name = experiment_name
        self._dataset_name = dataset_name

        self._show_batch_level = show_batch_level
        self._show_batch_window_averages = show_batch_window_averages
        self._show_epoch_level = show_epoch_level

        self._metric_names = metric_names

        self._batch_log_period = batch_log_period

        self._flush_period = flush_period

        self._log_dir = log_dir

        self._tensorboard_options = tensorboard_options or {}

        self._writers = dict()

    def _setup(self):
        try:

            # Metrics writer is used for batch level and epoch level
            success = self._setup_writer_for(METRIC_WRITER)

            if self._show_batch_level and self._show_batch_window_averages:
                success &= self._setup_writer_for(WINDOW_AVERAGED_METRICS_WRITER)

        except Exception as e:
            _.log_exception(self._log, "An exception occurred setting up the Tensorboard callback", e)
            success = False

        self._valid = success

        return self._valid

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        return self._setup()

    def on_batch_training_completed(self, dataset_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        current = self._get_logs_base(logs)
        global_iter = current['global_iter']

        if global_iter > 0 and (global_iter % self._batch_log_period != 0):
            return True

        success = True
        if self._show_batch_level:
            success &= self._write(current, global_iter, writer_type=METRIC_WRITER, batch_level=True)

        if self._show_batch_window_averages:
            success &= self._write(current, global_iter, writer_type=WINDOW_AVERAGED_METRICS_WRITER, batch_level=True)

        return success

    def on_epoch_completed(self, logs):

        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._show_epoch_level:
            return

        current = self._get_logs_base(logs)
        epoch = current['epoch']

        return self._write(current, epoch, writer_type=METRIC_WRITER, batch_level=False)

    def on_training_ended(self, stopped_early, stopped_on_error, callback_calls_success):
        for writer in self._writers.values():
            writer.close()
        return True

    def _setup_tensorboard_log_dir(self, writer_type):
        paths = [self._log_dir]
        if self._experiment_name:
            paths.append(self._experiment_name)

        subpath = ""
        if self._dataset_name:
            subpath = f'{self._dataset_name}_'

        subpath += writer_type
        paths.append(subpath)

        tensorboard_log_dir = os.path.join(*paths)

        self._log.debug(f'Tensorboard directory for {writer_type}: {tensorboard_log_dir}')

        if not os.path.exists(tensorboard_log_dir):
            self._log.debug(f'Directory doesn\'t exist, creating directory')
            os.makedirs(tensorboard_log_dir)

        return tensorboard_log_dir

    def _setup_writer_for(self, writer_type):
        self._log.debug(f'Setting up Tensorboard writer for {writer_type} ...')

        tensorboard_log_dir = self._setup_tensorboard_log_dir(writer_type)

        tensorboard_options = {**{
            'log_dir': tensorboard_log_dir
        }, **self._tensorboard_options}

        self._writers[writer_type] = SummaryWriter(**tensorboard_options)

        return True

    def _write(self, current_logs, training_iter, writer_type, batch_level):
        metrics, success = self._get_metrics_from(current_logs, writer_type, batch_level)

        for tag, metric in metrics.items():
            self._writers[writer_type].add_scalar(tag, metric, global_step=training_iter)

        if training_iter % self._flush_period == 0:
            self._writers[writer_type].flush()

        return success

    def _get_metrics_from(self, current_logs, writer_type, batch_level):
        # Get all metrics available
        if self._dataset_name:
            alt_base_path = None
            if batch_level:
                if writer_type != WINDOW_AVERAGED_METRICS_WRITER:
                    base_path = f"{self._dataset_name}.batch"
                else:
                    base_path = f"{self._dataset_name}.window_average"
            else:
                base_path = f"{self._dataset_name}.dataset"

                # for epoch-level fall back on window averages
                alt_base_path = f"{self._dataset_name}.window_average"

            metrics, success = self._get_all_metrics_from(base_path, current_logs, batch_level)
            if alt_base_path is not None:
                self._log.warn(f'Trying alternate metrics path {alt_base_path} ...')
                alt_metrics, alt_success = self._get_all_metrics_from(alt_base_path, current_logs, batch_level)

                if alt_success:
                    metrics = metrics or {}
                    for metric_path, alt_metric in alt_metrics.items():
                        if metric_path in metrics:
                            continue

                        metrics[metric_path] = alt_metric

                    success = len(metrics) > 0

            return metrics, success
        else:
            self._log.error(f'Don\'t know how to get metrics to log without a dataset name or '
                            f'a list of metrics to show')
            return None, False

    def _get_all_metrics_from(self, base_path, current_logs, batch_level):
        dataset_metrics = get_value_at(base_path, current_logs)

        # TODO : Make this a library level constant
        skip_metric_names = {"auxiliary_results"}

        metrics = {}

        def _add_metrics(m, base_path=None):
            for metric_name, metric in m.items():
                if metric_name in skip_metric_names:
                    continue

                if metric is None:
                    continue

                if base_path is not None:
                    metric_name = f"{base_path}.{metric_name}"

                if _.is_dict(metric):
                    _add_metrics(metric, metric_name)
                    continue

                tag = self._get_tag(metric_name, batch_level)

                metrics[tag] = metric

        if _.is_dict(dataset_metrics):
            _add_metrics(dataset_metrics)

            if len(metrics) > 0:
                return metrics, True
            else:
                self._log.warn(f'No metrics found at {base_path}')
                return metrics, False
        else:
            self._log.error(f'No valid metrics found for {base_path}')
            return None, False

    def _get_tag(self, metric_name, batch_level):
        prefix = 'batch' if batch_level else 'epoch'

        tag = self._metric_names[metric_name] if (metric_name in self._metric_names) else metric_name
        tag = tag.replace('.', '_')
        tag = f"{prefix}_{tag}"

        return tag



import os
import time
import json

import abc

from shutil import copyfile

from mlpug.utils import get_value_at
from mlpug.trainers.callbacks.callback import Callback

import basics.base_utils as _


class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):

    def __init__(self,
                 model_hyper_parameters=None,
                 checkpoints_path="../trained-models/",
                 batch_level=True,
                 metric_to_monitor="validation.window_average.perplexity",
                 metric_opt_mode='min',
                 metric_monitor_period=200,  # batches or epochs
                 create_checkpoint_every=200,  # batches or epochs
                 archive_last_model_checkpoint_every=2000,  # batches or epochs
                 force_monitoring_on_epoch=True,
                 base_checkpoint_filename=time.strftime("%d-%m-%Y_%H-%M-%S"),
                 model_checkpoint_filename_ext="model",
                 training_checkpoint_filename_ext="state",
                 backup_before_override=True,
                 disable_saving_checkpoints=False,
                 name="CheckpointManager"):
        """

        :param model_hyper_parameters: Dict with hyper parameters of the model, this will be saved as part of
                                       the model checkpoint
        :param checkpoints_path:
        :param batch_level: True if monitoring happens on the scale of batches. If False, monitoring is done
                            on the epoch level
        :param metric_to_monitor: key path to metric value in the log object, e.g. `validation.batch.perplexity`
        :param metric_opt_mode: 'max', 'min'
        :param metric_monitor_period: The period between checks for model quality improvement.
                                      This is in number of batches if `batch_level = True`, else it is a
                                      number of epochs.
        :param create_checkpoint_every: period before saving next training/model checkpoint
                                        (will be overridden after each period)
        :param archive_last_model_checkpoint_every: period before the last available model checkpoint is archived.
                                                    Period must be multiple of create_checkpoint_every period
        :param force_monitoring_on_epoch: When True, the given metric will also be monitored on every epoch
                                          in the case that monitoring level is batch level
        :param base_checkpoint_filename:
        :param model_checkpoint_filename_ext:
        :param training_checkpoint_filename_ext:
        :param backup_before_override: When True the training and model checkpoints are backed up before they are
                                       overridden by a new version. If backing up fails, no new version will be written
                                       to disk. This gives the user a chance to fix a problem, e.g. disk full, without
                                       interruption of the training process.
        :param disable_saving_checkpoints: When True no checkpoints at all will be saved, e.g. for testing training

        """
        super().__init__(name=name)

        self._model_hyper_parameters = model_hyper_parameters

        self._checkpoints_path = checkpoints_path
        self._model_checkpoint_filename_ext = model_checkpoint_filename_ext
        self._training_checkpoint_filename_ext = training_checkpoint_filename_ext
        self._base_filename = base_checkpoint_filename

        self._batch_level = batch_level

        self._metric_to_monitor = metric_to_monitor
        self._metric_opt_mode = metric_opt_mode

        self._metric_monitor_period = metric_monitor_period
        self._create_checkpoint_every = create_checkpoint_every
        self._archive_last_model_checkpoint_every = archive_last_model_checkpoint_every

        self._disable_saving_checkpoints = disable_saving_checkpoints

        self._force_monitoring_on_epoch = force_monitoring_on_epoch

        self._best_model_quality = float('Inf') if self._metric_opt_mode == 'min' else -float('Inf')
        self._best_model_iter = None

        self._backup_before_override = backup_before_override

        self._check_settings()

        self._describe_setup()

    def get_state(self):
        return {
            "best_model_quality": self._best_model_quality,
            "best_model_iter": self._best_model_iter
        }, True

    def set_state(self, state):
        """

        :param state Dict with saved checkpoint state to continue tracking the best model during training
        :return:
        """
        success = True
        if not _.is_dict(state):
            self._log.debug("No initial checkpoint state given, starting with clean slate ...")
            return success

        self._log.debug("Using given initial checkpoint state: \n\n%s" % json.dumps(state, indent=4))

        try:
            self._best_model_quality = state['best_model_quality']
            self._best_model_iter = state['best_model_iter']
        except Exception as e:
            _.log_exception(self._log, "Unable to set checkpoint manager state", e)
            success = False

        return success

    def on_batch_training_completed(self, training_batch, logs):
        if not self._batch_level:
            return True

        return self._monitor('global_iter', logs)

    def on_epoch_completed(self, logs):
        force = False
        iter_name = 'epoch'

        force_monitoring = False
        if self._batch_level:
            if self._force_monitoring_on_epoch:
                self._log.debug("Epoch completed : checking if model improved and creating checkpoints ... ")
                iter_name = 'global_iter'
                force_monitoring = True
            else:
                return True

        return self._monitor(iter_name, logs, force_monitoring=force_monitoring)

    def on_training_ended(self, stopped_early, stopped_on_error, callback_calls_success):
        if abs(self._best_model_quality) == float('Inf') or self._best_model_iter is None:
            return True

        self._log.info(f"Best model quality reached {self._metric_to_monitor}={self._best_model_quality} "
                       f"at global iteration {self._best_model_iter}")
        return True

    def training_checkpoint_file_name(self):
        b_fn = self._base_filename
        ext = self._training_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-training-checkpoint.{ext}')

    def current_model_file_name(self, training_iter):
        b_fn = self._base_filename
        ext = self._model_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-model-checkpoint-{training_iter}.{ext}')

    def latest_model_file_name(self):
        b_fn = self._base_filename
        ext = self._model_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-latest-model-checkpoint.{ext}')

    def best_model_file_name(self):
        b_fn = self._base_filename
        ext = self._model_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-best-model-checkpoint.{ext}')

    def _describe_setup(self):
        self._log.info(f"Metric to monitor : {self._metric_to_monitor}")
        self._log.info(f"Metric monitor period : {self._metric_monitor_period}")

        time_scale = 'batches' if self._batch_level else 'epochs'
        self._log.info(f"Create last training & model checkpoints every "
                       f"{self._archive_last_model_checkpoint_every} {time_scale}")
        self._log.info(f"Archive last model checkpoint every "
                       f"{self._archive_last_model_checkpoint_every} {time_scale}")

    def _get_model_quality(self, current_logs):
        return get_value_at(self._metric_to_monitor, current_logs)

    # TODO : it would maybe be better to split the monitoring and the checkpointing in to two separated methods
    def _monitor(self, iter_name, logs, force_monitoring=False):
        current = self._get_logs_base(logs)

        training_iter = current[iter_name]

        if not force_monitoring and training_iter == 0:
            return True

        best_model_fname = None
        success = True
        data_saved = False
        if force_monitoring or ((self._metric_monitor_period > 0) and (training_iter % self._metric_monitor_period == 0)):
            model_quality = self._get_model_quality(current)

            if model_quality is not None:
                model_improved = ((self._metric_opt_mode == 'min') and (model_quality < self._best_model_quality)) or \
                                 ((self._metric_opt_mode == 'max') and (model_quality > self._best_model_quality))

                if model_improved:
                    self._log.debug("Iter : %s : Model improved : %s : %3e " % (training_iter,
                                                                                self._metric_to_monitor,
                                                                                model_quality))

                    if (self._best_model_iter is not None) and \
                            (iter_name == 'epoch') and \
                            (training_iter < self._best_model_iter):
                        self._log.warn(f"Inconsistency: according to the current training iter ({training_iter}), "
                                       f"current best model training iter ({self._best_model_iter}) is in the future. "
                                       f"Was the right training checkpoint loaded?")

                    best_model_fname = self._disable_saving_checkpoints or self._save_current_model_as_best()
                    if best_model_fname:
                        self._best_model_quality = model_quality
                        self._best_model_iter = training_iter

                        data_saved = not self._disable_saving_checkpoints
                    else:
                        self._log.error("Unable to save improved model checkpoint")
                        success = False

                self._update_logs(model_improved, logs, current)
            else:
                self._log.error(f"No model quality available, unable to check if we need to save a checkpoint, "
                                f"skipping ...")
                success = False

        if not self._disable_saving_checkpoints and \
                (self._create_checkpoint_every > 0) and \
                (training_iter % self._create_checkpoint_every == 0):
            # Just copy best model if available
            latest_model_fname = self._create_model_checkpoint(file_to_copy=best_model_fname)

            latest_model_saved = (latest_model_fname is not None)
            success &= latest_model_saved

            data_saved |= latest_model_saved

            if (self._archive_last_model_checkpoint_every > 0) and \
                    (training_iter % self._archive_last_model_checkpoint_every == 0):
                if latest_model_saved:
                    copy_success = self._copy(latest_model_fname, self.current_model_file_name(training_iter))

                    success &= copy_success

                    data_saved |= copy_success
                else:
                    self._log.error("Unable to create checkpoint for latest model, unable to archive latest model")
                    success = False

            training_checkpoint_success = self._create_training_checkpoint()
            data_saved |= training_checkpoint_success

            success &= training_checkpoint_success

        if data_saved:
            self._log.debug("Saving of data done.\n\n")

        return success

    def _save_current_model_as_best(self):

        model_fn = self.best_model_file_name()

        if self._backup_before_override:
            try:
                self._backup_checkpoint(model_fn)
            except Exception as e:
                _.log_exception(self._log, f"A problem occurred backing up the last best model checkpoint, "
                                           f"will not override override model checkpoint with new one", e)
                return None

        try:
            self._save_model_checkpoint(model_fn)
        except Exception as e:
            _.log_exception(self._log, f"A problem occurred saving the current model as best model", e)
            return None

        return model_fn

    def _update_logs(self, model_improved, logs, current):
        if model_improved:
            logs["best"] = current.copy()

        current["is_best"] = model_improved

    def _create_model_checkpoint(self, file_to_copy=None):
        model_fn = self.latest_model_file_name()
        if file_to_copy:
            if not self._copy(file_to_copy, model_fn):
                self._log.error(f"Unable to create model checkpoint based on file : {file_to_copy}")
                return None
        else:
            try:
                self._save_model_checkpoint(model_fn)
            except Exception as e:
                _.log_exception(self._log, f"A problem occurred saving the latest model as checkpoint", e)
                return None

        return model_fn

    def _create_training_checkpoint(self):
        checkpoint_fname = self.training_checkpoint_file_name()

        if self._backup_before_override:
            try:
                if not self._backup_checkpoint(checkpoint_fname):
                    self._log.error("Unable to backup last training checkpoint, "
                                    "will not override override training checkpoint with new one")
                    return False
            except Exception as e:
                _.log_exception(self._log, f"A problem occurred backing up the last training checkpoint, "
                                           f"will not override override training checkpoint with new one", e)
                return False

        success = False
        try:
            success = self._save_training_checkpoint(checkpoint_fname)
        except Exception as e:
            _.log_exception(self._log, f"Unable to save training checkpoint", e)

        return success

    @abc.abstractmethod
    def _save_training_checkpoint(self, filename):
        self._log.error("This method is not implemented, implement it in your child class implementation")

    @abc.abstractmethod
    def _save_model_checkpoint(self, filename):
        self._log.error("This method is not implemented, implement it in your child class implementation")

    def _backup_checkpoint(self, filename):
        if os.path.isfile(filename):
            return self._copy(filename, f"{filename}.backup")
        else:
            return True

    def _copy(self, source_fname, dest_fname):
        try:
            self._log.debug("Copying model: [%s] ==> [%s]" % (source_fname, dest_fname))
            copyfile(source_fname, dest_fname)
        except Exception as e:
            _.log_exception(self._log, "Unable to copy [%s]" % source_fname, e)
            return False

        return True

    def _check_settings(self):
        if not os.path.exists(self._checkpoints_path):
            self._log.info(f"Creating checkpoint directory : {self._checkpoints_path}")
            os.makedirs(self._checkpoints_path)

        if (self._create_checkpoint_every < 0) and (self._archive_last_model_checkpoint_every > 0):
            self._log.error("archive_last_model_checkpoint_every can't be > 0 while _create_checkpoint_every < 0, "
                            "disabling archiving ... ")
            self._archive_last_model_checkpoint_every = -1

        if (self._create_checkpoint_every > 0) and \
           (self._archive_last_model_checkpoint_every > 0) and \
           (self._archive_last_model_checkpoint_every % self._create_checkpoint_every != 0):

            self._archive_last_model_checkpoint_every = 10 * self._create_checkpoint_every
            self._log.warn(f"archive_last_model_checkpoint_every must be "
                           f"exact multiple of _create_checkpoint_every, "
                           f"changing archive_last_model_checkpoint_every to "
                           f"[{self._archive_last_model_checkpoint_every}]")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mlpug",
    version="0.0.10",
    author="Freddy Snijder",
    author_email="mlpug@visionscapers.com",
    description="A machine learning library agnostic framework for model training",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nuhame/mlpug",
    packages=setuptools.find_packages(),
    install_requires=[
        'tensorboard',
        # required for tensorboard, else this error occurs : ModuleNotFoundError: No module named 'past'
        # TODO : still required?
        'future'
    ],
    dependency_links=['git+https://github.com/visionscaper/pybase.git'],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import torch
import torch.nn as nn

import torch.nn.functional as F

# from torchnlp.nn import LockedDropout


class EncoderRNN(nn.Module):
    def __init__(self, hidden_size, embedding, n_layers=1, dropout=0, enforce_sorted=True):
        super(EncoderRNN, self).__init__()
        self.n_layers = n_layers
        self.hidden_size = hidden_size
        self.embedding = embedding
        self.enforce_sorted = enforce_sorted

        # Initialize GRU; the input_size and hidden_size params are both set to 'hidden_size'
        #   because our input size is a word embedding with number of features == hidden_size
        self.gru = nn.GRU(embedding.embedding_dim, hidden_size, n_layers,
                          dropout=(0 if n_layers == 1 else dropout),
                          bidirectional=True)

    def forward(self, input_seq, input_lengths, hidden=None, return_state=True):
        total_length = input_seq.size(0)  # get the max sequence length

        # Convert word indexes to embeddings
        embedded = self.embedding(input_seq)

        # Batch dimension should be second (for partitioning over multiple GPUs)
        # But pack_padded_sequence requires a 1D list
        input_lengths = input_lengths.squeeze(0)
        # Pack padded batch of sequences for RNN module
        packed = nn.utils.rnn.pack_padded_sequence(embedded, input_lengths, enforce_sorted=self.enforce_sorted)

        # Forward pass through GRU
        self.gru.flatten_parameters()
        outputs, hidden = self.gru(packed, hidden)

        # Unpack padding
        outputs, _ = nn.utils.rnn.pad_packed_sequence(outputs, total_length=total_length)
        # Sum bidirectional GRU outputs
        outputs = outputs[:, :, :self.hidden_size] + outputs[:, :, self.hidden_size:]

        if return_state:
            # Return output and final hidden state
            return outputs, hidden
        else:
            return outputs


class Attn(nn.Module):
    def __init__(self, method, hidden_size):
        super(Attn, self).__init__()
        self.method = method
        if self.method not in ['dot', 'general', 'concat']:
            raise ValueError(self.method, "is not an appropriate attention method.")
        self.hidden_size = hidden_size
        if self.method == 'general':
            self.attn = nn.Linear(self.hidden_size, hidden_size)
        elif self.method == 'concat':
            self.attn = nn.Linear(self.hidden_size * 2, hidden_size)
            self.v = nn.Parameter(torch.FloatTensor(hidden_size))

    def dot_score(self, hidden, encoder_output):
        return torch.sum(hidden * encoder_output, dim=2)

    def general_score(self, hidden, encoder_output):
        energy = self.attn(encoder_output)
        return torch.sum(hidden * energy, dim=2)

    def concat_score(self, hidden, encoder_output):
        energy = self.attn(torch.cat((hidden.expand(encoder_output.size(0), -1, -1), encoder_output), 2)).tanh()
        return torch.sum(self.v * energy, dim=2)

    def forward(self, hidden, encoder_outputs):
        # Calculate the attention weights (energies) based on the given method
        if self.method == 'general':
            attn_energies = self.general_score(hidden, encoder_outputs)
        elif self.method == 'concat':
            attn_energies = self.concat_score(hidden, encoder_outputs)
        elif self.method == 'dot':
            attn_energies = self.dot_score(hidden, encoder_outputs)

        # Transpose max_length and batch_size dimensions
        attn_energies = attn_energies.t()

        # Return the softmax normalized probability scores (with added dimension)
        return F.softmax(attn_energies, dim=1).unsqueeze(1)


class LuongAttnDecoderRNN(nn.Module):
    def __init__(self, attn_model, embedding, hidden_size, output_size, n_layers=1, dropout=0.1):
        super(LuongAttnDecoderRNN, self).__init__()

        # Keep for reference
        self.attn_model = attn_model
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.n_layers = n_layers
        self.dropout = dropout

        # Define layers
        self.embedding = embedding
        self.embedding_dropout = nn.Dropout(dropout)
        self.gru = nn.GRU(embedding.embedding_dim,
                          hidden_size,
                          n_layers,
                          dropout=(0 if n_layers == 1 else dropout))

        self.concat = nn.Linear(hidden_size * 2, hidden_size)
        self.out = nn.Linear(hidden_size, output_size)

        self.attn = Attn(attn_model, hidden_size)

    def forward(self, input_step, last_hidden, encoder_outputs):
        # Note: we run this one step (word) at a time
        # Get embedding of current input word
        embedded = self.embedding(input_step)
        embedded = self.embedding_dropout(embedded)
        # Forward through unidirectional GRU
        self.gru.flatten_parameters()
        rnn_output, hidden = self.gru(embedded, last_hidden)
        # Calculate attention weights from the current GRU output
        attn_weights = self.attn(rnn_output, encoder_outputs)
        # Multiply attention weights to encoder outputs to get new "weighted sum" context vector
        context = attn_weights.bmm(encoder_outputs.transpose(0, 1))
        # Concatenate weighted context vector and GRU output using Luong eq. 5
        rnn_output = rnn_output.squeeze(0)
        context = context.squeeze(1)
        concat_input = torch.cat((rnn_output, context), 1)
        concat_output = torch.tanh(self.concat(concat_input))
        # Predict next word using Luong eq. 6
        output = self.out(concat_output)
        output = F.softmax(output, dim=1)

        # Ensure that batch dimension is the second dimension for proper gathering from multiple GPUs
        output = output.unsqueeze(0)
        # Return output and final hidden state
        return output, hidden



"""
Parsing functions for the Cornell movie dialogs corpus
"""

import os
import re
import unicodedata
import csv
import codecs

from mlpug.reference.chatbot_tutorial.voc import Voc

MOVIE_LINES_FIELDS = ["lineID", "characterID", "movieID", "character", "text"]
MOVIE_CONVERSATIONS_FIELDS = ["character1ID", "character2ID", "movieID", "utteranceIDs"]


def printLines(file, n=10):
    with open(file, 'rb') as datafile:
        lines = datafile.readlines()
    for line in lines[:n]:
        print(line)


# Splits each line of the file into a dictionary of fields
def loadLines(fileName, fields):
    lines = {}
    with open(fileName, 'r', encoding='iso-8859-1') as f:
        for line in f:
            values = line.split(" +++$+++ ")
            # Extract fields
            lineObj = {}
            for i, field in enumerate(fields):
                lineObj[field] = values[i]
            lines[lineObj['lineID']] = lineObj
    return lines


# Groups fields of lines from `loadLines` into conversations based on *movie_conversations.txt*
def loadConversations(fileName, lines, fields):
    conversations = []
    with open(fileName, 'r', encoding='iso-8859-1') as f:
        for line in f:
            values = line.split(" +++$+++ ")
            # Extract fields
            convObj = {}
            for i, field in enumerate(fields):
                convObj[field] = values[i]
            # Convert string to list (convObj["utteranceIDs"] == "['L598485', 'L598486', ...]")
            utterance_id_pattern = re.compile('L[0-9]+')
            lineIds = utterance_id_pattern.findall(convObj["utteranceIDs"])
            # Reassemble lines
            convObj["lines"] = []
            for lineId in lineIds:
                convObj["lines"].append(lines[lineId])
            conversations.append(convObj)
    return conversations


# Extracts pairs of sentences from conversations
def extractSentencePairs(conversations):
    qa_pairs = []
    for conversation in conversations:
        # Iterate over all the lines of the conversation
        for i in range(len(conversation["lines"]) - 1):  # We ignore the last line (no answer for it)
            inputLine = conversation["lines"][i]["text"].strip()
            targetLine = conversation["lines"][i+1]["text"].strip()
            # Filter wrong samples (if one of the lists is empty)
            if inputLine and targetLine:
                qa_pairs.append([inputLine, targetLine])
    return qa_pairs


# Turn a Unicode string to plain ASCII, thanks to
# https://stackoverflow.com/a/518232/2809427
def unicodeToAscii(s):
    return ''.join(
        c for c in unicodedata.normalize('NFD', s)
        if unicodedata.category(c) != 'Mn'
    )


# Lowercase, trim, and remove non-letter characters
def normalizeString(s):
    s = unicodeToAscii(s.lower().strip())
    s = re.sub(r"([.!?])", r" \1", s)
    s = re.sub(r"[^a-zA-Z.!?]+", r" ", s)
    s = re.sub(r"\s+", r" ", s).strip()
    return s


# Read query/response pairs
def readSentencePairs(datafile):
    print("Reading lines...")
    # Read the file and split into lines
    lines = open(datafile, encoding='utf-8').\
        read().strip().split('\n')
    # Split every line into pairs and normalize
    pairs = [[normalizeString(s) for s in l.split('\t')] for l in lines]
    return pairs


# Returns True iff both sentences in a pair 'p' are under the MAX_LENGTH threshold
def filterPair(p, max_length):
    # Input sequences need to preserve the last word for EOS token
    return len(p[0].split(' ')) < max_length and len(p[1].split(' ')) < max_length


# Filter pairs using filterPair condition
def filterPairs(pairs, max_length):
    return [pair for pair in pairs if filterPair(pair, max_length)]


def hasEmptyLine(p):
    return len(p[0]) == 0 or len(p[1]) == 0


def filterEmptyLines(pairs):
    return [pair for pair in pairs if not hasEmptyLine(pair)]

def normalize_sentences(pairs):
    return [[normalizeString(sentence) for sentence in pair] for pair in pairs]


def create_vocabulary(corpus_name, pairs, PAD_token, SOS_token, EOS_token):
    print("Counting words...")
    voc = Voc(corpus_name, PAD_token, SOS_token, EOS_token)
    for pair in pairs:
        voc.addSentence(pair[0])
        voc.addSentence(pair[1])
    print("Counted words:", voc.num_words)
    return voc


# Using the functions defined above, return a populated voc object and pairs list
def loadPrepareData(corpus_name, sentence_pairs_path, max_length, PAD_token, SOS_token, EOS_token):
    print("Start preparing training data ...")
    pairs = readSentencePairs(sentence_pairs_path)
    print("Read {!s} sentence pairs".format(len(pairs)))

    if max_length is not None:
        pairs = filterPairs(pairs, max_length)
        print("Trimmed to {!s} sentence pairs".format(len(pairs)))
    else:
        print("Skipped filtering of sentence pairs.")

    voc = create_vocabulary(corpus_name, pairs, PAD_token, SOS_token, EOS_token)
    return voc, pairs


def process_corpus_and_extract_sentence_pairs(corpus_path, output_path, output_delimiter):
    # Unescape the delimiter
    output_delimiter = str(codecs.decode(output_delimiter, "unicode_escape"))

    # Initialize lines dict, conversations list, and field ids
    lines = {}
    conversations = []

    # Load lines and process conversations
    print("\nProcessing corpus...")
    lines = loadLines(os.path.join(corpus_path, "movie_lines.txt"), MOVIE_LINES_FIELDS)
    print("\nLoading conversations...")
    conversations = loadConversations(os.path.join(corpus_path, "movie_conversations.txt"),
                                      lines, MOVIE_CONVERSATIONS_FIELDS)

    # Write new csv file
    print(f"\nWriting sentence pairs from conversations to  file: {output_path}")
    with open(output_path, 'w', encoding='utf-8') as output_file:
        writer = csv.writer(output_file, delimiter=output_delimiter, lineterminator='\n')
        for pair in extractSentencePairs(conversations):
            writer.writerow(pair)


def trimRareWords(voc, pairs, MIN_COUNT):
    # Trim words used under the MIN_COUNT from the voc
    voc.trim(MIN_COUNT)
    # Filter out pairs with trimmed words
    keep_pairs = []
    for pair in pairs:
        input_sentence = pair[0]
        output_sentence = pair[1]
        keep_input = True
        keep_output = True
        # Check input sentence
        for word in input_sentence.split(' '):
            if word not in voc.word2index:
                keep_input = False
                break
        # Check output sentence
        for word in output_sentence.split(' '):
            if word not in voc.word2index:
                keep_output = False
                break

        # Only keep pairs that do not contain trimmed word(s) in their input or output sentence
        if keep_input and keep_output:
            keep_pairs.append(pair)

    print("Trimmed from {} pairs to {}, {:.4f} of total".format(len(pairs), len(keep_pairs), len(keep_pairs) / len(pairs)))
    return keep_pairs


